"""
File with the available transforms for images
"""

import torchvision.transforms
import sys
import unittest


def __getattr__(name):
    """
    When a function is called for the heat.vision_transforms module it will attempt to run the
    heat function/class with that name, then, if there is no such object, it will attempt to
    get the torchvision.transforms object of that name.
    """
    if name in torchvision.transforms.transforms.__all__:
        return torchvision.transforms.__getattribute__(name)
    else:
        raise AttributeError(f"module {name} not implemented in Torch or Heat")



"""
Collection of solvers for systems of linear equations.
"""
import heat as ht
from ..dndarray import DNDarray
from ..sanitation import sanitize_out
from typing import List, Dict, Any, TypeVar, Union, Tuple, Optional

import torch

__all__ = ["cg", "lanczos"]


def cg(A: DNDarray, b: DNDarray, x0: DNDarray, out: Optional[DNDarray] = None) -> DNDarray:
    """
    Conjugate gradients method for solving a system of linear equations :math: `Ax = b`

    Parameters
    ----------
    A : DNDarray
        2D symmetric, positive definite Matrix
    b : DNDarray
        1D vector
    x0 : DNDarray
        Arbitrary 1D starting vector
    out : DNDarray, optional
        Output Vector
    """
    if not isinstance(A, DNDarray) or not isinstance(b, DNDarray) or not isinstance(x0, DNDarray):
        raise TypeError(
            f"A, b and x0 need to be of type ht.DNDarray, but were {type(A)}, {type(b)}, {type(x0)}"
        )

    if A.ndim != 2:
        raise RuntimeError("A needs to be a 2D matrix")
    if b.ndim != 1:
        raise RuntimeError("b needs to be a 1D vector")
    if x0.ndim != 1:
        raise RuntimeError("c needs to be a 1D vector")

    r = b - ht.matmul(A, x0)
    p = r
    rsold = ht.matmul(r, r)
    x = x0

    for i in range(len(b)):
        Ap = ht.matmul(A, p)
        alpha = rsold / ht.matmul(p, Ap)
        x = x + alpha * p
        r = r - alpha * Ap
        rsnew = ht.matmul(r, r)
        if ht.sqrt(rsnew).item() < 1e-10:
            print(f"Residual reaches tolerance in it = {i}")
            if out is not None:
                out = x
                return out
            return x
        p = r + ((rsnew / rsold) * p)
        rsold = rsnew

    if out is not None:
        out = x
        return out
    return x


def lanczos(
    A: DNDarray,
    m: int,
    v0: Optional[DNDarray] = None,
    V_out: Optional[DNDarray] = None,
    T_out: Optional[DNDarray] = None,
) -> Tuple[DNDarray, DNDarray]:
    r"""
    The Lanczos algorithm is an iterative approximation of the solution to the eigenvalue problem, as an adaptation of
    power methods to find the m "most useful" (tending towards extreme highest/lowest) eigenvalues and eigenvectors of
    an :math:`n \times n` Hermitian matrix, where often :math:`m<<n`.
    It returns two matrices :math:`V` and :math:`T`, where:

        - :math:`V` is a Matrix of size :math:`n\times m`, with orthonormal columns, that span the Krylow subspace \n
        - :math:`T` is a Tridiagonal matrix of size :math:`m\times m`, with coefficients :math:`\alpha_1,..., \alpha_n`
          on the diagonal and coefficients :math:`\beta_1,...,\beta_{n-1}` on the side-diagonals\n

    Parameters
    ----------
    A : DNDarray
        2D Hermitian (if complex) or symmetric positive-definite matrix.
        Only distribution along axis 0 is supported, i.e. `A.split` must be `0` or `None`.
    m : int
        Number of Lanczos iterations
    v0 : DNDarray, optional
        1D starting vector of Euclidean norm 1. If not provided, a random vector will be used to start the algorithm
    V_out : DNDarray, optional
        Output Matrix for the Krylow vectors, Shape = (n, m), dtype=A.dtype, must be initialized to zero
    T_out : DNDarray, optional
        Output Matrix for the Tridiagonal matrix, Shape = (m, m), must be initialized to zero
    """
    if not isinstance(A, DNDarray):
        raise TypeError(f"A needs to be of type ht.dndarray, but was {type(A)}")
    if A.ndim != 2:
        raise RuntimeError("A needs to be a 2D matrix")
    if A.dtype is ht.int32 or A.dtype is ht.int64:
        raise TypeError(f"A can be float or complex, got {A.dtype}")
    if not isinstance(m, (int, float)):
        raise TypeError(f"m must be int, got {type(m)}")

    n, column = A.shape
    if n != column:
        raise TypeError("Input Matrix A needs to be symmetric positive-definite.")

    # output data types: T is always Real
    A_is_complex = A.dtype is ht.complex128 or A.dtype is ht.complex64
    T_dtype = A.real.dtype

    # initialize or sanitize output buffers
    if T_out is not None:
        sanitize_out(
            T_out,
            output_shape=(m, m),
            output_split=None,
            output_device=A.device,
            output_comm=A.comm,
        )
        T = T_out
    else:
        T = ht.zeros((m, m), dtype=T_dtype, device=A.device, comm=A.comm)
    if A.split == 0:
        if V_out is not None:
            sanitize_out(
                V_out,
                output_shape=(n, m),
                output_split=0,
                output_device=A.device,
                output_comm=A.comm,
            )
            V = V_out
        else:
            # This is done for better memory access in the reorthogonalization Gram-Schmidt algorithm
            V = ht.zeros((n, m), split=0, dtype=A.dtype, device=A.device, comm=A.comm)
    else:
        if A.split == 1:
            raise NotImplementedError("Distribution along axis 1 not implemented yet.")
        if V_out is not None:
            sanitize_out(
                V_out,
                output_shape=(n, m),
                output_split=None,
                output_device=A.device,
                output_comm=A.comm,
            )
            V = V_out
        else:
            V = ht.zeros((n, m), split=None, dtype=A.dtype, device=A.device, comm=A.comm)

    if A_is_complex:
        if v0 is None:
            vr = (
                ht.random.rand(n, split=V.split, dtype=T_dtype, device=V.device, comm=V.comm)
                + ht.random.rand(n, split=V.split, dtype=T_dtype, device=V.device, comm=V.comm) * 1j
            )
            v0 = vr / ht.norm(vr)
        elif v0.split != V.split:
            v0.resplit_(axis=V.split)
        # # 0th iteration
        # # vector v0 has Euclidean norm = 1
        w = ht.matmul(A, v0)
        alpha = ht.dot(ht.conj(w).T, v0)
        w = w - alpha * v0
        T[0, 0] = alpha.real
        V[:, 0] = v0
        for i in range(1, int(m)):
            beta = ht.norm(w)
            if ht.abs(beta) < 1e-10:
                # print("Lanczos breakdown in iteration {}".format(i))
                # Lanczos Breakdown, pick a random vector to continue
                vr = (
                    ht.random.rand(n, split=V.split, dtype=T_dtype, device=V.device, comm=V.comm)
                    + ht.random.rand(n, split=V.split, dtype=T_dtype, device=V.device, comm=V.comm)
                    * 1j
                )
                # orthogonalize v_r with respect to all vectors v[i]
                for j in range(i):
                    vi_loc = V._DNDarray__array[:, j]
                    a = torch.dot(vr.larray, torch.conj(vi_loc))
                    b = torch.dot(vi_loc, torch.conj(vi_loc))
                    A.comm.Allreduce(ht.communication.MPI.IN_PLACE, a, ht.communication.MPI.SUM)
                    A.comm.Allreduce(ht.communication.MPI.IN_PLACE, b, ht.communication.MPI.SUM)
                    vr._DNDarray__array = vr._DNDarray__array - a / b * vi_loc
                # normalize v_r to Euclidean norm 1 and set as ith vector v
                vi = vr / ht.norm(vr)
            else:
                vr = w

                # Reorthogonalization
                for j in range(i):
                    vi_loc = V.larray[:, j]
                    a = torch.dot(vr._DNDarray__array, torch.conj(vi_loc))
                    b = torch.dot(vi_loc, torch.conj(vi_loc))
                    A.comm.Allreduce(ht.communication.MPI.IN_PLACE, a, ht.communication.MPI.SUM)
                    A.comm.Allreduce(ht.communication.MPI.IN_PLACE, b, ht.communication.MPI.SUM)
                    vr._DNDarray__array = vr._DNDarray__array - a / b * vi_loc

                vi = vr / ht.norm(vr)

            w = ht.matmul(A, vi)
            alpha = ht.dot(ht.conj(w).T, vi)

            w = w - alpha * vi - beta * V[:, i - 1]

            T[i - 1, i] = beta.real
            T[i, i - 1] = beta.real
            T[i, i] = alpha.real
            V[:, i] = vi
    else:
        if v0 is None:
            vr = ht.random.rand(n, split=V.split, dtype=T_dtype, device=V.device, comm=V.comm)
            v0 = vr / ht.norm(vr)
        elif v0.split != V.split:
            v0.resplit_(axis=V.split)
        # # 0th iteration
        # # vector v0 has Euclidean norm = 1
        w = ht.matmul(A, v0)
        alpha = ht.dot(w, v0)
        w = w - alpha * v0
        T[0, 0] = alpha
        V[:, 0] = v0
        for i in range(1, int(m)):
            beta = ht.norm(w)
            if ht.abs(beta) < 1e-10:
                # print("Lanczos breakdown in iteration {}".format(i))
                # Lanczos Breakdown, pick a random vector to continue
                vr = ht.random.rand(n, split=V.split, dtype=T_dtype, device=V.device, comm=V.comm)
                # orthogonalize v_r with respect to all vectors v[i]
                for j in range(i):
                    vi_loc = V._DNDarray__array[:, j]
                    a = torch.dot(vr.larray, vi_loc)
                    b = torch.dot(vi_loc, vi_loc)
                    A.comm.Allreduce(ht.communication.MPI.IN_PLACE, a, ht.communication.MPI.SUM)
                    A.comm.Allreduce(ht.communication.MPI.IN_PLACE, b, ht.communication.MPI.SUM)
                    vr._DNDarray__array = vr._DNDarray__array - a / b * vi_loc
                # normalize v_r to Euclidean norm 1 and set as ith vector v
                vi = vr / ht.norm(vr)
            else:
                vr = w

                # Reorthogonalization
                for j in range(i):
                    vi_loc = V.larray[:, j]
                    a = torch.dot(vr._DNDarray__array, vi_loc)
                    b = torch.dot(vi_loc, vi_loc)
                    A.comm.Allreduce(ht.communication.MPI.IN_PLACE, a, ht.communication.MPI.SUM)
                    A.comm.Allreduce(ht.communication.MPI.IN_PLACE, b, ht.communication.MPI.SUM)
                    vr._DNDarray__array = vr._DNDarray__array - a / b * vi_loc

                vi = vr / ht.norm(vr)

            w = ht.matmul(A, vi)
            alpha = ht.dot(w, vi)

            w = w - alpha * vi - beta * V[:, i - 1]

            T[i - 1, i] = beta
            T[i, i - 1] = beta
            T[i, i] = alpha
            V[:, i] = vi

    if V.split is not None:
        V.resplit_(axis=None)

    return V, T



"""
A collection of functions used for inferring or correcting things before major computation
"""

import itertools
import numpy as np
import torch

from typing import Tuple, Union


def broadcast_shape(shape_a: Tuple[int, ...], shape_b: Tuple[int, ...]) -> Tuple[int, ...]:
    """
    Infers, if possible, the broadcast output shape of two operands a and b. Inspired by stackoverflow post:
    https://stackoverflow.com/questions/24743753/test-if-an-array-is-broadcastable-to-a-shape

    Parameters
    ----------
    shape_a : Tuple[int,...]
        Shape of first operand
    shape_b : Tuple[int,...]
        Shape of second operand

    Raises
    -------
    ValueError
        If the two shapes cannot be broadcast.

    Examples
    --------
    >>> import heat as ht
    >>> ht.core.stride_tricks.broadcast_shape((5,4),(4,))
    (5, 4)
    >>> ht.core.stride_tricks.broadcast_shape((1,100,1),(10,1,5))
    (10, 100, 5)
    >>> ht.core.stride_tricks.broadcast_shape((8,1,6,1),(7,1,5,))
    (8,7,6,5))
    >>> ht.core.stride_tricks.broadcast_shape((2,1),(8,4,3))
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      File "heat/core/stride_tricks.py", line 42, in broadcast_shape
        "operands could not be broadcast, input shapes {} {}".format(shape_a, shape_b)
    ValueError: operands could not be broadcast, input shapes (2, 1) (8, 4, 3)
    """
    try:
        resulting_shape = torch.broadcast_shapes(shape_a, shape_b)
    except AttributeError:  # torch < 1.8
        it = itertools.zip_longest(shape_a[::-1], shape_b[::-1], fillvalue=1)
        resulting_shape = max(len(shape_a), len(shape_b)) * [None]
        for i, (a, b) in enumerate(it):
            if a == 0 and b == 1 or b == 0 and a == 1:
                resulting_shape[i] = 0
            elif a == 1 or b == 1 or a == b:
                resulting_shape[i] = max(a, b)
            else:
                raise ValueError(
                    f"operands could not be broadcast, input shapes {shape_a} {shape_b}"
                )
        return tuple(resulting_shape[::-1])
    except TypeError:
        raise TypeError(f"operand 1 must be tuple of ints, not {type(shape_a)}")
    except NameError:
        raise TypeError(f"operands must be tuples of ints, not {shape_a} and {shape_b}")
    except RuntimeError:
        raise ValueError(f"operands could not be broadcast, input shapes {shape_a} {shape_b}")

    return tuple(resulting_shape)


def sanitize_axis(
    shape: Tuple[int, ...], axis: Union[int, None, Tuple[int, ...]]
) -> Union[int, None, Tuple[int, ...]]:
    """
    Checks conformity of an axis with respect to a given shape. The axis will be converted to its positive equivalent
    and is checked to be within bounds

    Parameters
    ----------
    shape : Tuple[int, ...]
        Shape of an array
    axis : ints or Tuple[int, ...] or None
        The axis to be sanitized

    Raises
    -------
    ValueError
        if the axis cannot be sanitized, i.e. out of bounds.
    TypeError
        if the axis is not integral.

    Examples
    -------
    >>> import heat as ht
    >>> ht.core.stride_tricks.sanitize_axis((5,4,4),1)
    1
    >>> ht.core.stride_tricks.sanitize_axis((5,4,4),-1)
    2
    >>> ht.core.stride_tricks.sanitize_axis((5, 4), (1,))
    (1,)
    >>> ht.core.stride_tricks.sanitize_axis((5, 4), 1.0)
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      File "heat/heat/core/stride_tricks.py", line 99, in sanitize_axis
        raise TypeError("axis must be None or int or tuple, but was {}".format(type(axis)))
    TypeError: axis must be None or int or tuple, but was <class 'float'>

    """
    # scalars are handled like unsplit matrices
    if len(shape) == 0:
        axis = None

    if axis is not None and not isinstance(axis, int) and not isinstance(axis, tuple):
        raise TypeError(f"axis must be None or int or tuple, but was {type(axis)}")
    if isinstance(axis, tuple):
        axis = tuple(dim + len(shape) if dim < 0 else dim for dim in axis)
        for dim in axis:
            if dim < 0 or dim >= len(shape):
                raise ValueError(f"axis {axis} is out of bounds for shape {shape}")
        return axis

    if axis is None or 0 <= axis < len(shape):
        return axis
    elif axis < 0:
        axis += len(shape)

    if axis < 0 or axis >= len(shape):
        raise ValueError(f"axis {axis} is out of bounds for shape {shape}")

    return axis


def sanitize_shape(shape: Union[int, Tuple[int, ...]], lval: int = 0) -> Tuple[int, ...]:
    """
    Verifies and normalizes the given shape.

    Parameters
    ----------
    shape : int or Tupe[int,...]
        Shape of an array.
    lval : int
        Lowest legal value

    Raises
    -------
    ValueError
        If the shape contains illegal values, e.g. negative numbers.
    TypeError
        If the given shape is neither and int or a sequence of ints.

    Examples
    --------
    >>> import heat as ht
    >>> ht.core.stride_tricks.sanitize_shape(3)
    (3,)
    >>> ht.core.stride_tricks.sanitize_shape([1, 2, 3])
    (1, 2, 3,)
    >>> ht.core.stride_tricks.sanitize_shape(1.0)
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      File "heat/heat/core/stride_tricks.py", line 159, in sanitize_shape
        raise TypeError("expected sequence object with length >= 0 or a single integer")
    TypeError: expected sequence object with length >= 0 or a single integer
    """
    shape = tuple(shape) if hasattr(shape, "__iter__") else (shape,)

    for dimension in shape:
        if issubclass(type(dimension), np.integer):
            dimension = int(dimension)
        if not isinstance(dimension, int):
            raise TypeError("expected sequence object with length >= 0 or a single integer")
        if dimension < lval:
            raise ValueError("negative dimensions are not allowed")

    return shape


def sanitize_slice(sl: slice, max_dim: int) -> slice:
    """
    Remove None-types from a slice

    Parameters
    ----------
    sl : slice
        slice to adjust
    max_dim : int
        maximum index for the given slice

    Raises
    ------
    TypeError
        if sl is not a slice.
    """
    if not isinstance(sl, slice):
        raise TypeError("This function is only for slices!")

    new_sl = [None] * 3
    new_sl[0] = 0 if sl.start is None else sl.start
    if new_sl[0] < 0:
        new_sl[0] += max_dim

    new_sl[1] = max_dim if sl.stop is None else sl.stop
    if new_sl[1] < 0:
        new_sl[1] += max_dim

    new_sl[2] = 1 if sl.step is None else sl.step

    return slice(new_sl[0], new_sl[1], new_sl[2])



"""
handle different devices. Current options: CPU (default), GPU
"""

from __future__ import annotations

import torch

from typing import Any, Optional, Union

from . import communication


__all__ = ["Device", "cpu", "get_device", "sanitize_device", "use_device"]


class Device:
    """
    Implements a compute device. HeAT can run computations on different compute devices or backends.
    A device describes the device type and id on which said computation should be carried out.

    Parameters
    ----------
    device_type : str
        Represents HeAT's device name
    device_id : int
        The device id
    torch_device : str
        The corresponding PyTorch device type

    Examples
    --------
    >>> ht.Device("cpu", 0, "cpu:0")
    device(cpu:0)
    >>> ht.Device("gpu", 0, "cuda:0")
    device(gpu:0)
    """

    def __init__(self, device_type: str, device_id: int, torch_device: str):
        self.__device_type = device_type
        self.__device_id = device_id
        self.__torch_device = torch_device

    @property
    def device_type(self) -> str:
        """
        Return the type of :class:`~heat.core.device.Device` as a string.
        """
        return self.__device_type

    @property
    def device_id(self) -> int:
        """
        Return the identification number of :class:`~heat.core.device.Device`.
        """
        return self.__device_id

    @property
    def torch_device(self) -> str:
        """
        Return the type and id of :class:`~heat.core.device.Device` as a PyTorch device string object.
        """
        return self.__torch_device

    def __repr__(self) -> str:
        """
        Return the unambiguous information of :class:`~heat.core.device.Device`.
        """
        return f"device({self.__str__()})"

    def __str__(self) -> str:
        """
        Return the descriptive information of :class:`~heat.core.device.Device`.
        """
        return f"{self.device_type}:{self.device_id}"

    def __eq__(self, other: Any) -> bool:
        """
        Overloads the `==` operator for local equal check.

        Parameters
        ----------
        other : Any
            The object to compare with
        """
        if isinstance(other, Device):
            return self.device_type == other.device_type and self.device_id == other.device_id
        elif isinstance(other, torch.device):
            return self.device_type == other.type and self.device_id == other.index
        else:
            return NotImplemented


# create a CPU device singleton
cpu = Device("cpu", 0, "cpu")
"""
The standard CPU Device

Examples
--------
>>> ht.cpu
device(cpu:0)
>>> ht.ones((2, 3), device=ht.cpu)
DNDarray([[1., 1., 1.],
          [1., 1., 1.]], dtype=ht.float32, device=cpu:0, split=None)
"""

# define the default device to be the CPU
__default_device = cpu
# add a device string for the CPU device
__device_mapping = {cpu.device_type: cpu}

# add gpu support if available
if torch.cuda.device_count() > 0:
    # GPUs are assigned round-robin to the MPI processes
    gpu_id = communication.MPI_WORLD.rank % torch.cuda.device_count()
    # create a new GPU device
    gpu = Device("gpu", gpu_id, f"cuda:{gpu_id}")
    """
    The standard GPU Device

    Examples
    --------
    >>> ht.cpu
    device(cpu:0)
    >>> ht.ones((2, 3), device=ht.gpu)
    DNDarray([[1., 1., 1.],
          [1., 1., 1.]], dtype=ht.float32, device=gpu:0, split=None)
    """
    # add a GPU device string
    __device_mapping[gpu.device_type] = gpu
    __device_mapping["cuda"] = gpu
    # the GPU device should be exported as global symbol
    __all__.append("gpu")


def get_device() -> Device:
    """
    Retrieves the currently globally set default :class:`~heat.core.device.Device`.
    """
    return __default_device


def sanitize_device(device: Optional[Union[str, Device]] = None) -> Device:
    """
    Sanitizes a device or device identifier, i.e. checks whether it is already an instance of :class:`~heat.core.device.Device` or a string with
    known device identifier and maps it to a proper :class:`~heat.core.device.Device`.

    Parameters
    ----------
    device : str or Device, optional
        The device to be sanitized

    Raises
    ------
    ValueError
        If the given device id is not recognized
    """
    if device is None:
        return get_device()

    if isinstance(device, Device):
        return device

    try:
        return __device_mapping[device.strip().lower()]
    except (AttributeError, KeyError, TypeError):
        raise ValueError(f'Unknown device, must be one of {", ".join(__device_mapping.keys())}')


def use_device(device: Optional[Union[str, Device]] = None) -> None:
    """
    Sets the globally used default :class:`~heat.core.device.Device`.

    Parameters
    ----------
    device : str or Device
        The device to be set
    """
    global __default_device
    __default_device = sanitize_device(device)



"""
Utility functions for the heat optimizers
"""

import math
import torch

from typing import Optional, Dict


__all__ = ["DetectMetricPlateau"]


class DetectMetricPlateau(object):
    r"""
    Determine if a  when a metric has stopped improving.
    This scheduler reads a metrics quantity and if no improvement
    is seen for a 'patience' number of epochs, the learning rate is reduced.

    Adapted from `torch.optim.lr_scheduler.ReduceLROnPlateau <https://pytorch.org/docs/stable/optim.html#torch.optim.lr_scheduler.ReduceLROnPlateau>`_.

    Args:
        mode: str, optional
            One of `min`, `max`.
            In `min` mode, the quantity monitored is determined to have plateaued when
            it stops decreasing. In `max` mode, the quantity monitored is determined to
            have plateaued when it stops decreasing.\n
            Default: 'min'.
        patience: int, optional
            Number of epochs to wait before determining if there is a plateau
            For example, if `patience = 2`, then we will ignore the first 2 epochs
            with no improvement, and will only determine if there is a plateau after the
            3rd epoch if the loss still hasn't improved then.\n
            Default: 10.
        threshold: float, optional
            Threshold for measuring the new optimum to only focus on significant changes.\n
            Default: 1e-4.
        threshold_mode: str, optional
            One of `rel`, `abs`. In `rel` mode,
            dynamic_threshold = best * ( 1 + threshold ) in 'max'
            mode or best * ( 1 - threshold ) in `min` mode.
            In `abs` mode, dynamic_threshold = best + threshold in
            `max` mode or best - threshold in `min` mode.\n
            Default: 'rel'.
        cooldown: int, optional
            Number of epochs to wait before resuming
            normal operation after lr has been reduced.\n
            Default: 0.
    """

    def __init__(
        self,
        mode: Optional[str] = "min",
        patience: Optional[int] = 10,
        threshold: Optional[float] = 1e-4,
        threshold_mode: Optional[str] = "rel",
        cooldown: Optional[int] = 0,
    ):  # noqa: D107
        self.patience = patience
        self.cooldown = cooldown
        self.cooldown_counter = 0
        self.mode = mode
        self.threshold = threshold
        self.threshold_mode = threshold_mode
        self.best = None
        self.num_bad_epochs = None
        self.mode_worse = None  # the worse value for the chosen mode
        self.last_epoch = 0
        self._init_is_better(mode=mode, threshold=threshold, threshold_mode=threshold_mode)
        self.reset()

    def get_state(self) -> Dict:
        """
        Get a dictionary of the class parameters. This is useful for checkpointing.
        """
        return {
            "patience": self.patience,
            "cooldown": self.cooldown,
            "cooldown_counter": self.cooldown_counter,
            "mode": self.mode,
            "threshold": self.threshold,
            "threshold_mode": self.threshold_mode,
            "best": self.best,
            "num_bad_epochs": self.num_bad_epochs,
            "mode_worse": self.mode_worse,
            "last_epoch": self.last_epoch,
        }

    def set_state(self, dic: Dict) -> None:
        """
        Load a dictionary with the status of the class. Typically used in checkpointing.

        Parameters
        ----------
        dic: Dictionary
            contains the values to be set as the class parameters
        """
        self.patience = dic["patience"]
        self.cooldown = dic["cooldown"]
        self.cooldown_counter = dic["cooldown_counter"]
        self.mode = dic["mode"]
        self.threshold = dic["threshold"]
        self.threshold_mode = dic["threshold_mode"]
        self.best = dic["best"]
        self.num_bad_epochs = dic["num_bad_epochs"]
        self.mode_worse = dic["mode_worse"]
        self.last_epoch = dic["last_epoch"]

    def reset(self) -> None:
        """
        Resets num_bad_epochs counter and cooldown counter.
        """
        self.best = self.mode_worse
        self.cooldown_counter = 0
        self.num_bad_epochs = 0

    def test_if_improving(self, metrics: torch.Tensor) -> bool:
        """
        Test if the metric/s is/are improving. If the metrics are better than the adjusted best value, they
        are set as the best for future testing.

        Parameters
        ----------
        metrics: torch.Tensor
            the metrics to test

        Returns
        -------
        True if the metrics are better than the best, False otherwise
        """
        # convert `metrics` to float, in case it's a zero-dim Tensor
        current = float(metrics)

        epoch = self.last_epoch + 1
        self.last_epoch = epoch

        if self.is_better(current, self.best):
            self.best = current
            self.num_bad_epochs = 0
        else:
            self.num_bad_epochs += 1

        if self.in_cooldown:
            self.cooldown_counter -= 1
            self.num_bad_epochs = 0  # ignore any bad epochs in cooldown

        if self.num_bad_epochs > self.patience:
            self.cooldown_counter = self.cooldown
            self.num_bad_epochs = 0
            return True
        return False

    @property
    def in_cooldown(self) -> bool:
        """
        Test if the class is in the cool down period
        """
        return self.cooldown_counter > 0

    def is_better(self, a: float, best: float) -> bool:
        """
        Test if the given value is better than the current best value. The best value is adjusted with the threshold

        Parameters
        ----------
        a: float
            the metric value
        best: float
            the current best value for the metric

        Returns
        -------
        boolean indicating if the metric is improving
        """
        if self.mode == "min" and self.threshold_mode == "rel":
            rel_epsilon = 1.0 - self.threshold
            comp = best * rel_epsilon if best >= 0 else best * (1 + self.threshold)
            return a < comp

        elif self.mode == "min" and self.threshold_mode == "abs":
            return a < best - self.threshold

        elif self.mode == "max" and self.threshold_mode == "rel":
            rel_epsilon = self.threshold + 1.0
            return a > best * rel_epsilon

        else:  # mode == 'max' and epsilon_mode == 'abs':
            return a > best + self.threshold

    def _init_is_better(self, mode: str, threshold: float, threshold_mode: str) -> None:
        """
        Initialize the is_better function for comparisons later
        """
        if mode not in {"min", "max"}:
            raise ValueError("mode " + mode + " is unknown!")
        if threshold_mode not in {"rel", "abs"}:
            raise ValueError("threshold mode " + threshold_mode + " is unknown!")

        if mode == "min":
            self.mode_worse = math.inf
        else:  # mode == 'max':
            self.mode_worse = -math.inf

        self.mode = mode
        self.threshold = threshold
        self.threshold_mode = threshold_mode



"""
Collection of validation/sanitation routines.
"""
from __future__ import annotations

import numpy as np
import torch
import warnings
from typing import Any, Union, Sequence, List, Tuple

from .communication import MPI, Communication
from .dndarray import DNDarray

from . import factories
from . import stride_tricks
from . import types


__all__ = [
    "sanitize_distribution",
    "sanitize_in",
    "sanitize_infinity",
    "sanitize_in_tensor",
    "sanitize_lshape",
    "sanitize_out",
    "sanitize_sequence",
    "scalar_to_1d",
]


def sanitize_distribution(
    *args: DNDarray, target: DNDarray, diff_map: torch.Tensor = None
) -> Union[DNDarray, Tuple(DNDarray)]:
    """
    Distribute every `arg` according to `target.lshape_map` or, if provided, `diff_map`.
    After this sanitation, the lshapes are compatible along the split dimension.
    `Args` can contain non-distributed DNDarrays, they will be split afterwards, if `target` is split.

    Parameters
    ----------
    args : DNDarray
        Dndarrays to be distributed

    target : DNDarray
        Dndarray used to sanitize the metadata and to, if diff_map is not given, determine the resulting distribution.

    diff_map : torch.Tensor (optional)
        Different lshape_map. Overwrites the distribution of the target array.
        Used in cases when the target array does not correspond to the actually wanted distribution,
        e.g. because it only contains a single element along the split axis and gets broadcast.

    Raises
    ------
    TypeError
        When an argument is not a ``DNDarray`` or ``None``.
    ValueError
        When the split-axes or sizes along the split-axis do not match.

    See Also
    ---------
    :func:`~heat.core.dndarray.create_lshape_map`
        Function to create the lshape_map.
    """
    out = []
    sanitize_in(target)
    # early out on 1 process
    if target.comm.size == 1:
        for arg in args:
            sanitize_in(arg)
            out.append(arg)
        return tuple(out) if len(out) > 1 else out[0]

    target_split = target.split
    if diff_map is not None:
        sanitize_in_tensor(diff_map)
        target_map = diff_map
        if target_split is not None:
            tmap_split = target_map[:, target_split]
            target_size = tmap_split.sum().item()
            # Check if the diff_map is balanced
            w_size = target_map.shape[0]
            tmap_balanced = torch.full_like(tmap_split, fill_value=target_size // w_size)
            remainder = target_size % w_size
            tmap_balanced[:remainder] += 1
            target_balanced = torch.equal(tmap_balanced, tmap_split)
    elif target_split is not None:
        target_map = target.lshape_map
        target_size = target.shape[target_split]
        target_balanced = target.is_balanced(force_check=False)

    for arg in args:
        sanitize_in(arg)
        if target.comm != arg.comm:
            try:
                raise NotImplementedError(
                    f"Not implemented for other comms, found {target.comm.name} and {arg.comm.name}"
                )
            except Exception:
                raise NotImplementedError("Not implemented for other comms")
        elif target_split is None:
            if arg.split is not None:
                raise NotImplementedError(
                    f"DNDarrays must have the same split axes, found {target_split} and {arg.split}"
                )
            else:
                out.append(arg)
        elif arg.shape[target_split] == 1 and target_size > 1:  # broadcasting in split-dimension
            out.append(arg.resplit(None))
        elif arg.shape[target_split] != target_size:
            raise ValueError(
                f"Cannot distribute to match in split dimension, shapes are {target.shape} and {arg.shape}"
            )
        elif arg.split is None:  # undistributed case
            if target_balanced:
                out.append(
                    factories.array(
                        arg, split=target_split, copy=False, comm=arg.comm, device=arg.device
                    )
                )
            else:
                idx = [slice(None)] * arg.ndim
                idx[target_split] = slice(
                    target_map[: arg.comm.rank, target_split].sum(),
                    target_map[: arg.comm.rank + 1, target_split].sum(),
                )
                out.append(
                    factories.array(
                        arg.larray[tuple(idx)],
                        is_split=target_split,
                        copy=False,
                        comm=arg.comm,
                        device=arg.device,
                    )
                )
        elif arg.split != target_split:
            raise NotImplementedError(
                f"DNDarrays must have the same split axes, found {target_split} and {arg.split}"
            )
        elif not (
            # False
            target_balanced
            and arg.is_balanced(force_check=False)
        ):  # Split axes are the same and atleast one is not balanced
            current_map = arg.lshape_map
            out_map = current_map.clone()
            out_map[:, target_split] = target_map[:, target_split]
            if not (current_map[:, target_split] == target_map[:, target_split]).all():
                out.append(arg.redistribute(lshape_map=current_map, target_map=out_map))
            else:
                out.append(arg)
        else:  # both are balanced
            out.append(arg)
    if len(out) == 1:
        return out[0]
    return tuple(out)


def sanitize_in(x: Any):
    """
    Verify that input object is ``DNDarray``.

    Parameters
    ----------
    x : Any
        Input object

    Raises
    ------
    TypeError
        When ``x`` is not a ``DNDarray``.
    """
    if not isinstance(x, DNDarray):
        raise TypeError(f"Input must be a DNDarray, is {type(x)}")


def sanitize_infinity(x: Union[DNDarray, torch.Tensor]) -> Union[int, float]:
    """
    Returns largest possible value for the ``dtype`` of the input array.

    Parameters
    -----------
    x: Union[DNDarray, torch.Tensor]
        Input object.
    """
    dtype = x.dtype if isinstance(x, torch.Tensor) else x.larray.dtype
    try:
        largest = torch.finfo(dtype).max
    except TypeError:
        largest = torch.iinfo(dtype).max

    return largest


def sanitize_in_tensor(x: Any):
    """
    Verify that input object is ``torch.Tensor``.

    Parameters
    ----------
    x : Any
        Input object.

    Raises
    ------
    TypeError
        When ``x`` is not a ``torch.Tensor``.
    """
    if not isinstance(x, torch.Tensor):
        raise TypeError(f"Input must be a torch.Tensor, is {type(x)}")


def sanitize_lshape(array: DNDarray, tensor: torch.Tensor):
    """
    Verify shape consistency when manipulating process-local arrays.

    Parameters
    ----------
    array : DNDarray
        the original, potentially distributed ``DNDarray``
    tensor : torch.Tensor
        process-local data meant to replace ``array.larray``

    Raises
    ------
    ValueError
        if shape of local ``torch.Tensor`` is inconsistent with global ``DNDarray``.
    """
    # input sanitation is done in parent function
    tshape = tuple(tensor.shape)
    if tshape == array.lshape:
        return
    gshape = array.gshape
    split = array.split
    if split is None:
        # allow for axes with size 0, other axes must match
        non_zero = [i for i in range(len(tshape)) if tshape[i] != 0]
        cond = [tshape[i] == gshape[i] for i in non_zero].count(True) == len(non_zero)
        if cond:
            return
        else:
            raise ValueError(
                f"Shape of local tensor is inconsistent with global DNDarray: tensor.shape is {tshape}, should be {gshape}"
            )
    # size of non-split dimensions must match global shape
    reduced_gshape = gshape[:split] + gshape[split + 1 :]
    reduced_tshape = tshape[:split] + tshape[split + 1 :]
    if reduced_tshape == reduced_gshape:
        return
    raise ValueError(
        f"Shape of local tensor along non-split axes is inconsistent with global DNDarray: tensor.shape is {tshape}, DNDarray is {gshape}"
    )


def sanitize_out(
    out: Any,
    output_shape: Tuple,
    output_split: int,
    output_device: str,
    output_comm: Communication = None,
):
    """
    Validate output buffer ``out``.

    Parameters
    ----------
    out : Any
          the `out` buffer where the result of some operation will be stored

    output_shape : Tuple
                   the calculated shape returned by the operation

    output_split : Int
                   the calculated split axis returned by the operation

    output_device : Str
                    "cpu" or "gpu" as per location of data

    output_comm : Communication
                    Communication object of the result of the operation

    Raises
    ------
    TypeError
        if ``out`` is not a ``DNDarray``.
    ValueError
        if shape, split direction, or device of the output buffer ``out`` do not match the operation result.
    """
    if not isinstance(out, DNDarray):
        raise TypeError(f"expected `out` to be None or a DNDarray, but was {type(out)}")

    out_proxy = out.__torch_proxy__()
    out_proxy.names = [
        "split" if (out.split is not None and i == out.split) else f"_{i}"
        for i in range(out_proxy.ndim)
    ]
    out_proxy = out_proxy.squeeze()

    check_proxy = torch.ones(1).expand(output_shape)
    check_proxy.names = [
        "split" if (output_split is not None and i == output_split) else f"_{i}"
        for i in range(check_proxy.ndim)
    ]
    check_proxy = check_proxy.squeeze()

    if out_proxy.shape != check_proxy.shape:
        raise ValueError(f"Expecting output buffer of shape {output_shape}, got {out.shape}")
    count_split = int(out.split is not None) + int(output_split is not None)
    if count_split == 1:
        raise ValueError(
            "Split axis of output buffer is inconsistent with split semantics for this operation."
        )
    elif count_split == 2:
        if out.shape[out.split] > 1:  # split axis is not squeezed out
            if out_proxy.names.index("split") != check_proxy.names.index("split"):
                raise ValueError(
                    "Split axis of output buffer is inconsistent with split semantics for this operation."
                )
        else:  # split axis is squeezed out
            num_dim_before_split = len(
                [name for name in out_proxy.names if int(name[1:]) < out.split]
            )
            check_num_dim_before_split = len(
                [name for name in check_proxy.names if int(name[1:]) < output_split]
            )
            if num_dim_before_split != check_num_dim_before_split:
                raise ValueError(
                    "Split axis of output buffer is inconsistent with split semantics for this operation."
                )
    if out.device != output_device:
        raise ValueError(f"Device mismatch: out is on {out.device}, should be on {output_device}")
    if output_comm is not None and out.comm != output_comm:
        try:
            raise NotImplementedError(
                f"Not implemented for other comms, found {out.comm.name} and {output_comm.name}"
            )
        except Exception:
            raise NotImplementedError("Not implemented for other comms")


def sanitize_sequence(
    seq: Union[Sequence[int, ...], Sequence[float, ...], DNDarray, torch.Tensor]
) -> List:
    """
    Check if sequence is valid, return list.

    Parameters
    ----------
    seq : Union[Sequence[int, ...], Sequence[float, ...], DNDarray, torch.Tensor]
        Input sequence.

    Raises
    ------
    TypeError
        if ``seq`` is neither a list nor a tuple
    """
    if isinstance(seq, list):
        return seq
    elif isinstance(seq, tuple):
        return list(seq)
    else:
        raise TypeError(f"seq must be a list or a tuple, got {type(seq)}")


def scalar_to_1d(x: DNDarray) -> DNDarray:
    """
    Turn a scalar ``DNDarray`` into a 1-D ``DNDarray`` with 1 element.

    Parameters
    ----------
    x : DNDarray
        with `x.ndim = 0`
    """
    return factories.array(
        x.larray.unsqueeze(0), dtype=x.dtype, split=x.split, comm=x.comm, device=x.device
    )



"""Provides mixins for high-level algorithms, e.g. classifiers or clustering algorithms."""

import inspect
import json

from typing import Dict, List, TypeVar

from .dndarray import DNDarray

self = TypeVar("self")


class BaseEstimator:
    """
    Abstract base class for all estimators, i.e. parametrized analysis algorithms, in HeAT. Can be used as mixin.
    """

    @classmethod
    def _parameter_names(cls) -> List[str]:
        """
        Get the names of all parameters that can be set inside the constructor of the estimator.
        """
        init = cls.__init__
        if init is object.__init__:
            return []

        # introspect the constructor arguments to find the model parameters
        init_signature = inspect.signature(init)

        # consider the constructor parameters excluding 'self'
        return [
            p.name
            for p in init_signature.parameters.values()
            if p.name != "self" and p.kind == p.POSITIONAL_OR_KEYWORD
        ]

    def get_params(self, deep: bool = True) -> Dict[str, object]:
        """
        Get parameters for this estimator.

        Parameters
        ----------
        deep : bool, default: True
            If ``True``, will return the parameters for this estimator and contained sub-objects that are estimators.
        """
        params = {}

        for key in self._parameter_names():
            value = getattr(self, key)

            if deep and hasattr(value, "get_params"):
                value = value.get_params()

            params[key] = value
        return params

    def __repr__(self, indent: int = 1) -> str:
        """
        Returns a printable representation of the object.

        Parameters
        ----------
        indent : int, default: 1
            Indicates the indentation for the top-level output.
        """
        return "{}({})".format(self.__class__.__name__, json.dumps(self.get_params(), indent=4))

    def set_params(self, **params: Dict[str, object]) -> self:
        """
        Set the parameters of this estimator. The method works on simple estimators as well as on nested objects
        (such as pipelines). The latter have to be nested dictionaries.

        Parameters
        ----------
        **params : dict[str, object]
            Estimator parameters to bet set.
        """
        if not params:
            return self

        parameter_names = self._parameter_names()
        for key, value in params.items():
            if key not in parameter_names:
                raise ValueError(
                    f"Invalid parameter {key} for estimator {self}. Check the list of available parameters with `estimator.get_params().keys()`."
                )

            if isinstance(value, dict):
                getattr(self, key).set_params(value)
            else:
                setattr(self, key, value)

        return self


class ClassificationMixin:
    """
    Mixin for all classifiers in HeAT.
    """

    def fit(self, x: DNDarray, y: DNDarray):
        """
        Fits the classification model.

        Parameters
        ----------
        x : DNDarray
            Training instances to train on. Shape = (n_samples, n_features)

        y : DNDarray
            Class values to fit. Shape = (n_samples, )

        """
        raise NotImplementedError()

    def fit_predict(self, x: DNDarray, y: DNDarray) -> DNDarray:
        """
        Fits model and returns classes for each input sample
        Convenience method; equivalent to calling :func:`fit` followed by :func:`predict`.

        Parameters
        ----------
        x : DNDarray
            Input data to be predicted. Shape = (n_samples, n_features)
        y : DNDarray
            Class values to fit. Shape = (n_samples, )
        """
        self.fit(x, y)
        return self.predict(x)

    def predict(self, x: DNDarray) -> DNDarray:
        """
        Predicts the class labels for each sample.

        Parameters
        ----------
        x : DNDarray
            Values to predict the classes for. Shape = (n_samples, n_features)
        """
        raise NotImplementedError()


class ClusteringMixin:
    """
    Clustering mixin for all clusterers in HeAT.
    """

    def fit(self, x: DNDarray):
        """
        Computes the clustering.

        Parameters
        ----------
        x : DNDarray
            Training instances to cluster. Shape = (n_samples, n_features)
        """
        raise NotImplementedError()

    def fit_predict(self, x: DNDarray) -> DNDarray:
        """
        Compute clusters and returns the predicted cluster assignment for each sample.
        Returns index of the cluster each sample belongs to.
        Convenience method; equivalent to calling :func:`fit` followed by :func:`predict`.

        Parameters
        ----------
        x : DNDarray
            Input data to be clustered. Shape = (n_samples, n_features)
        """
        self.fit(x)
        return self.predict(x)


class RegressionMixin:
    """
    Mixin for all regression estimators in HeAT.
    """

    def fit(self, x: DNDarray, y: DNDarray):
        """
        Fits the regression model.

        Parameters
        ----------
        x : DNDarray
            Training instances to train on. Shape = (n_samples, n_features)
        y : DNDarray
            Continuous values to fit. Shape = (n_samples,)
        """
        raise NotImplementedError()

    def fit_predict(self, x: DNDarray, y: DNDarray) -> DNDarray:
        """
        Fits model and returns regression predictions for each input sample
        Convenience method; equivalent to calling :func:`fit` followed by :func:`predict`.

        Parameters
        ----------
        x : DNDarray
            Input data to be predicted. Shape = (n_samples, n_features)
        y : DNDarray
            Continuous values to fit. Shape = (n_samples,)
        """
        self.fit(x, y)
        return self.predict(x)

    def predict(self, x: DNDarray) -> DNDarray:
        """
        Predicts the continuous labels for each sample.

        Parameters
        ----------
        x : DNDarray
            Values to let the model predict. Shape = (n_samples, n_features)
        """
        raise NotImplementedError()


def is_classifier(estimator: object) -> bool:
    """
    Return ``True`` if the given estimator is a classifier, ``False`` otherwise.

    Parameters
    ----------
    estimator : object
        Estimator object to test.
    """
    return isinstance(estimator, ClassificationMixin)


def is_estimator(estimator: object) -> bool:
    """
    Return ``True`` if the given estimator is an estimator, ``False`` otherwise.

    Parameters
    ----------
    estimator : object
        Estimator object to test.
    """
    return isinstance(estimator, BaseEstimator)


def is_clusterer(estimator: object) -> bool:
    """
    Return ``True`` if the given estimator is a clusterer, ``False`` otherwise.

    Parameters
    ----------
    estimator : object
        Estimator object to test.

    """
    return isinstance(estimator, ClusteringMixin)


def is_regressor(estimator: object) -> bool:
    """
    Return ``True`` if the given estimator is a regressor, ``False`` otherwise.

    Parameters
    ----------
    estimator : object
        Estimator object to test.
    """
    return isinstance(estimator, RegressionMixin)



from setuptools import setup, find_packages
import codecs


with codecs.open("README.md", "r", "utf-8") as handle:
    long_description = handle.read()

__version__ = None  # appeases flake, assignment in exec() below
with open("./heat/core/version.py") as handle:
    exec(handle.read())

setup(
    name="heat",
    packages=find_packages(exclude=("*tests*", "*benchmarks*")),
    package_data={"heat.datasets": ["*.csv", "*.h5", "*.nc"]},
    version=__version__,
    description="A framework for high-performance data analytics and machine learning.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Helmholtz Association",
    author_email="martin.siggel@dlr.de",
    url="https://github.com/helmholtz-analytics/heat",
    keywords=["data", "analytics", "tensors", "distributed", "gpu"],
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
    ],
    install_requires=[
        "mpi4py>=3.0.0",
        "numpy>=1.20.0",
        "torch>=1.8.0, <2.0.2",
        "scipy>=0.14.0",
        "pillow>=6.0.0",
        "torchvision>=0.8.0",
    ],
    extras_require={
        "docutils": ["docutils>=0.16"],
        "hdf5": ["h5py>=2.8.0"],
        "netcdf": ["netCDF4>=1.5.6"],
        "dev": ["pre-commit>=1.18.3"],
        "examples": ["scikit-learn>=0.24.0", "matplotlib>=3.1.0"],
        "cb": ["perun>=0.2.0"],
    },
)



"""Enables parallel I/O with data on disk."""
from __future__ import annotations

import os.path
from math import log10
import numpy as np
import torch
import warnings

from typing import Dict, Iterable, List, Optional, Tuple, Union

from . import devices
from . import factories
from . import types

from .communication import Communication, MPI, MPI_WORLD, sanitize_comm
from .dndarray import DNDarray
from .manipulations import hsplit, vsplit
from .statistics import max as smax, min as smin
from .stride_tricks import sanitize_axis
from .types import datatype

__VALID_WRITE_MODES = frozenset(["w", "a", "r+"])
__CSV_EXTENSION = frozenset([".csv"])
__HDF5_EXTENSIONS = frozenset([".h5", ".hdf5"])
__NETCDF_EXTENSIONS = frozenset([".nc", ".nc4", "netcdf"])
__NETCDF_DIM_TEMPLATE = "{}_dim_{}"

__all__ = ["load", "load_csv", "save_csv", "save", "supports_hdf5", "supports_netcdf"]

try:
    import h5py
except ImportError:
    # HDF5 support is optional
    def supports_hdf5() -> bool:
        """
        Returns ``True`` if HeAT supports reading from and writing to HDF5 files, ``False`` otherwise.
        """
        return False

else:
    # add functions to exports
    __all__.extend(["load_hdf5", "save_hdf5"])

    # warn the user about serial hdf5
    if not h5py.get_config().mpi and MPI_WORLD.rank == 0:
        warnings.warn(
            "h5py does not support parallel I/O, falling back to slower serial I/O", ImportWarning
        )

    def supports_hdf5() -> bool:
        """
        Returns ``True`` if HeAT supports reading from and writing to HDF5 files, ``False`` otherwise.
        """
        return True

    def load_hdf5(
        path: str,
        dataset: str,
        dtype: datatype = types.float32,
        split: Optional[int] = None,
        device: Optional[str] = None,
        comm: Optional[Communication] = None,
    ) -> DNDarray:
        """
        Loads data from an HDF5 file. The data may be distributed among multiple processing nodes via the split flag.

        Parameters
        ----------
        path : str
            Path to the HDF5 file to be read.
        dataset : str
            Name of the dataset to be read.
        dtype : datatype, optional
            Data type of the resulting array.
        split : int or None, optional
            The axis along which the data is distributed among the processing cores.
        device : str, optional
            The device id on which to place the data, defaults to globally set default device.
        comm : Communication, optional
            The communication to use for the data distribution.

        Raises
        -------
        TypeError
            If any of the input parameters are not of correct type

        Examples
        --------
        >>> a = ht.load_hdf5('data.h5', dataset='DATA')
        >>> a.shape
        [0/2] (5,)
        [1/2] (5,)
        >>> a.lshape
        [0/2] (5,)
        [1/2] (5,)
        >>> b = ht.load_hdf5('data.h5', dataset='DATA', split=0)
        >>> b.shape
        [0/2] (5,)
        [1/2] (5,)
        >>> b.lshape
        [0/2] (3,)
        [1/2] (2,)
        """
        if not isinstance(path, str):
            raise TypeError(f"path must be str, not {type(path)}")
        elif not isinstance(dataset, str):
            raise TypeError("dataset must be str, not {}".format(type(dataset)))
        elif split is not None and not isinstance(split, int):
            raise TypeError(f"split must be None or int, not {type(split)}")

        # infer the type and communicator for the loaded array
        dtype = types.canonical_heat_type(dtype)
        # determine the comm and device the data will be placed on
        device = devices.sanitize_device(device)
        comm = sanitize_comm(comm)

        # actually load the data from the HDF5 file
        with h5py.File(path, "r") as handle:
            data = handle[dataset]
            gshape = tuple(data.shape)
            dims = len(gshape)
            split = sanitize_axis(gshape, split)
            _, _, indices = comm.chunk(gshape, split)
            balanced = True
            if split is None:
                data = torch.tensor(
                    data[indices], dtype=dtype.torch_type(), device=device.torch_device
                )
            elif indices[split].stop > indices[split].start:
                data = torch.tensor(
                    data[indices], dtype=dtype.torch_type(), device=device.torch_device
                )
            else:
                warnings.warn("More MPI ranks are used then the length of splitting dimension!")
                slice1 = tuple(
                    slice(0, gshape[i]) if i != split else slice(0, 1) for i in range(dims)
                )
                slice2 = tuple(
                    slice(0, gshape[i]) if i != split else slice(0, 0) for i in range(dims)
                )
                data = torch.tensor(
                    data[slice1], dtype=dtype.torch_type(), device=device.torch_device
                )
                data = data[slice2]

            return DNDarray(data, gshape, dtype, split, device, comm, balanced)

    def save_hdf5(
        data: DNDarray, path: str, dataset: str, mode: str = "w", **kwargs: Dict[str, object]
    ):
        """
        Saves ``data`` to an HDF5 file. Attempts to utilize parallel I/O if possible.

        Parameters
        ----------
        data : DNDarray
            The data to be saved on disk.
        path : str
            Path to the HDF5 file to be written.
        dataset : str
            Name of the dataset the data is saved to.
        mode : str, optional
            File access mode, one of ``'w', 'a', 'r+'``
        kwargs : dict, optional
            Additional arguments passed to the created dataset.

        Raises
        -------
        TypeError
            If any of the input parameters are not of correct type.
        ValueError
            If the access mode is not understood.

        Examples
        --------
        >>> x = ht.arange(100, split=0)
        >>> ht.save_hdf5(x, 'data.h5', dataset='DATA')
        """
        if not isinstance(data, DNDarray):
            raise TypeError(f"data must be heat tensor, not {type(data)}")
        if not isinstance(path, str):
            raise TypeError(f"path must be str, not {type(path)}")
        if not isinstance(dataset, str):
            raise TypeError(f"dataset must be str, not {type(path)}")

        # we only support a subset of possible modes
        if mode not in __VALID_WRITE_MODES:
            raise ValueError(f"mode was {mode}, not in possible modes {__VALID_WRITE_MODES}")

        # chunk the data, if no split is set maximize parallel I/O and chunk first axis
        is_split = data.split is not None
        _, _, slices = data.comm.chunk(data.gshape, data.split if is_split else 0)

        # attempt to perform parallel I/O if possible
        if h5py.get_config().mpi:
            with h5py.File(path, mode, driver="mpio", comm=data.comm.handle) as handle:
                dset = handle.create_dataset(dataset, data.shape, **kwargs)
                dset[slices] = data.larray.cpu() if is_split else data.larray[slices].cpu()

        # otherwise a single rank only write is performed in case of local data (i.e. no split)
        elif data.comm.rank == 0:
            with h5py.File(path, mode) as handle:
                dset = handle.create_dataset(dataset, data.shape, **kwargs)
                if is_split:
                    dset[slices] = data.larray.cpu()
                else:
                    dset[...] = data.larray.cpu()

            # ping next rank if it exists
            if is_split and data.comm.size > 1:
                data.comm.Isend([None, 0, MPI.INT], dest=1)
                data.comm.Recv([None, 0, MPI.INT], source=data.comm.size - 1)

        # no MPI, but split data is more tricky, we have to serialize the writes
        elif is_split:
            # wait for the previous rank to finish writing its chunk, then write own part
            data.comm.Recv([None, 0, MPI.INT], source=data.comm.rank - 1)
            with h5py.File(path, "r+") as handle:
                handle[dataset][slices] = data.larray.cpu()

            # ping the next node in the communicator, wrap around to 0 to complete barrier behavior
            next_rank = (data.comm.rank + 1) % data.comm.size
            data.comm.Isend([None, 0, MPI.INT], dest=next_rank)

    DNDarray.save_hdf5 = lambda self, path, dataset, mode="w", **kwargs: save_hdf5(
        self, path, dataset, mode, **kwargs
    )
    DNDarray.save_hdf5.__doc__ = save_hdf5.__doc__


try:
    import netCDF4 as nc
except ImportError:
    # netCDF4 support is optional
    def supports_netcdf() -> bool:
        """
        Returns ``True`` if HeAT supports reading from and writing to netCDF4 files, ``False`` otherwise.
        """
        return False

else:
    # add functions to visible exports
    __all__.extend(["load_netcdf", "save_netcdf"])

    # determine netCDF's parallel I/O support
    __nc_has_par = (
        nc.__dict__.get("__has_parallel4_support__", False)
        or nc.__dict__.get("__has_pnetcdf_support__", False)
        or nc.__dict__.get("__has_nc_par__", False)
    )

    # warn the user about serial netcdf
    if not __nc_has_par and MPI_WORLD.rank == 0:
        warnings.warn(
            "netCDF4 does not support parallel I/O, falling back to slower serial I/O",
            ImportWarning,
        )

    def supports_netcdf() -> bool:
        """
        Returns ``True`` if HeAT supports reading from and writing to netCDF4 files, ``False`` otherwise.
        """
        return True

    def load_netcdf(
        path: str,
        variable: str,
        dtype: datatype = types.float32,
        split: Optional[int] = None,
        device: Optional[str] = None,
        comm: Optional[Communication] = None,
    ) -> DNDarray:
        """
        Loads data from a NetCDF4 file. The data may be distributed among multiple processing nodes via the split flag.

        Parameters
        ----------
        path : str
            Path to the NetCDF4 file to be read.
        variable : str
            Name of the variable to be read.
        dtype : datatype, optional
            Data type of the resulting array
        split : int or None, optional
            The axis along which the data is distributed among the processing cores.
        comm : Communication, optional
            The communication to use for the data distribution. Defaults to MPI_COMM_WORLD.
        device : str, optional
            The device id on which to place the data, defaults to globally set default device.

        Raises
        -------
        TypeError
            If any of the input parameters are not of correct type.

        Examples
        --------
        >>> a = ht.load_netcdf('data.nc', variable='DATA')
        >>> a.shape
        [0/2] (5,)
        [1/2] (5,)
        >>> a.lshape
        [0/2] (5,)
        [1/2] (5,)
        >>> b = ht.load_netcdf('data.nc', variable='DATA', split=0)
        >>> b.shape
        [0/2] (5,)
        [1/2] (5,)
        >>> b.lshape
        [0/2] (3,)
        [1/2] (2,)
        """
        if not isinstance(path, str):
            raise TypeError(f"path must be str, not {type(path)}")
        if not isinstance(variable, str):
            raise TypeError(f"dataset must be str, not {type(variable)}")
        if split is not None and not isinstance(split, int):
            raise TypeError(f"split must be None or int, not {type(split)}")

        # infer the canonical heat datatype
        dtype = types.canonical_heat_type(dtype)
        # determine the device and comm the data will be placed on
        device = devices.sanitize_device(device)
        comm = sanitize_comm(comm)

        # actually load the data
        with nc.Dataset(path, "r", parallel=__nc_has_par, comm=comm.handle) as handle:
            data = handle[variable]

            # prepare meta information
            gshape = tuple(data.shape)
            split = sanitize_axis(gshape, split)

            # chunk up the data portion
            _, local_shape, indices = comm.chunk(gshape, split)
            balanced = True
            if split is None or local_shape[split] > 0:
                data = torch.tensor(
                    data[indices], dtype=dtype.torch_type(), device=device.torch_device
                )
            else:
                data = torch.empty(
                    local_shape, dtype=dtype.torch_type(), device=device.torch_device
                )

            return DNDarray(data, gshape, dtype, split, device, comm, balanced)

    def save_netcdf(
        data: DNDarray,
        path: str,
        variable: str,
        mode: str = "w",
        dimension_names: Union[list, tuple, str] = None,
        is_unlimited: bool = False,
        file_slices: Union[Iterable[int], slice, bool] = slice(None),
        **kwargs: Dict[str, object],
    ):
        """
        Saves data to a netCDF4 file. Attempts to utilize parallel I/O if possible.

        Parameters
        ----------
        data : DNDarray
            The data to be saved on disk.
        path : str
            Path to the netCDF4 file to be written.
        variable : str
            Name of the variable the data is saved to.
        mode : str, optional
            File access mode, one of ``'w', 'a', 'r+'``.
        dimension_names : list or tuple or string
            Specifies the netCDF Dimensions used by the variable. Ignored if Variable already exists.
        is_unlimited : bool, optional
            If True, every dimension created for this variable (i.e. doesn't already exist) is unlimited. Already
            existing limited dimensions cannot be changed to unlimited and vice versa.
        file_slices : integer iterable, slice, ellipsis or bool
            Keys used to slice the netCDF Variable, as given in the nc.utils._StartCountStride method.
        kwargs : dict, optional
            additional arguments passed to the created dataset.

        Raises
        -------
        TypeError
            If any of the input parameters are not of correct type.
        ValueError
            If the access mode is not understood or if the number of dimension names does not match the number of
            dimensions.

        Examples
        --------
        >>> x = ht.arange(100, split=0)
        >>> ht.save_netcdf(x, 'data.nc', dataset='DATA')
        """
        if not isinstance(data, DNDarray):
            raise TypeError("data must be heat tensor, not {}".format(type(data)))
        if not isinstance(path, str):
            raise TypeError("path must be str, not {}".format(type(path)))
        if not isinstance(variable, str):
            raise TypeError("variable must be str, not {}".format(type(path)))
        if dimension_names is None:
            dimension_names = [
                __NETCDF_DIM_TEMPLATE.format(variable, dim) for dim, _ in enumerate(data.shape)
            ]
        elif isinstance(dimension_names, str):
            dimension_names = [dimension_names]
        elif isinstance(dimension_names, tuple):
            dimension_names = list(dimension_names)
        elif not isinstance(dimension_names, list):
            raise TypeError(
                "dimension_names must be list or tuple or string, not{}".format(
                    type(dimension_names)
                )
            )
        elif not len(dimension_names) == len(data.shape):
            raise ValueError(
                "{0} names given for {1} dimensions".format(len(dimension_names), len(data.shape))
            )

        # we only support a subset of possible modes
        if mode not in __VALID_WRITE_MODES:
            raise ValueError(
                "mode was {}, not in possible modes {}".format(mode, __VALID_WRITE_MODES)
            )

        failed = 0
        excep = None
        # chunk the data, if no split is set maximize parallel I/O and chunk first axis
        is_split = data.split is not None
        _, _, slices = data.comm.chunk(data.gshape, data.split if is_split else 0)

        def __get_expanded_split(
            shape: Tuple[int], expanded_shape: Tuple[int], split: Optional[int]
        ) -> int:
            """
            Returns the hypothetical split-axis of a dndarray of shape=shape and
            split=split if it was expanded to expandedShape by adding empty dimensions.

            Parameters
            ----------
            shape : tuple[int]
                Shape of a DNDarray.
            expanded_shape : tuple[int]
                Shape of hypothetical expanded DNDarray.
            split : int or None
                split-axis of dndarray.

            Raises
            -------
            ValueError
                If resulting shapes do not match.
            """
            if np.prod(shape) != np.prod(expanded_shape):
                raise ValueError(
                    "Shapes %s and %s do not have the same size" % (shape, expanded_shape)
                )
            if np.prod(shape) == 1:  # size 1 array
                return split
            if len(shape) == len(expanded_shape):  # actually not expanded at all
                return split
            if split is None:  # not split at all
                return None
            # Get indices of non-empty dimensions and squeezed shapes
            enumerated = [[i, v] for i, v in enumerate(shape) if v != 1]
            ind_nonempty, sq_shape = list(zip(*enumerated))  # transpose
            enumerated = [[i, v] for i, v in enumerate(expanded_shape) if v != 1]
            ex_ind_nonempty, sq_ex = list(zip(*enumerated))  # transpose
            if not sq_shape == sq_ex:
                raise ValueError(
                    "Shapes %s and %s differ in non-empty dimensions" % (shape, expanded_shape)
                )
            if split in ind_nonempty:  # split along non-empty dimension
                split_sq = ind_nonempty.index(split)  # split-axis in squeezed shape
                return ex_ind_nonempty[split_sq]
            # split along empty dimension: split doesnt matter, only one process contains data
            # return the last empty dimension (in expanded shape) before (the first nonempty dimension after split)
            # number of nonempty elems before split
            ne_before_split = split - shape[:split].count(1)
            ind_ne_after_split = ind_nonempty[
                ne_before_split
            ]  # index of (first nonempty element after split) in squeezed shape
            return max(
                i
                for i, v in enumerate(expanded_shape[: max(ex_ind_nonempty[:ind_ne_after_split])])
                if v == 1
            )

        def __merge_slices(
            var: nc.Variable,
            var_slices: Tuple[int, slice],
            data: DNDarray,
            data_slices: Optional[Tuple[int, slice]] = None,
        ) -> Tuple[Union[int, slice]]:
            """
            This method allows replacing:
                ``var[var_slices][data_slices] = data``
            (a `netcdf4.Variable.__getitem__` and a `numpy.ndarray.__setitem__` call)

            with:
                ``var[ __merge_slices(var, var_slices, data, data_slices) ] = data``
            (a single `netcdf4.Variable.__setitem__` call).

            This is necessary because performing the former would, in the ``__getitem__``, load the global dataset onto
            every process in local ``np.ndarray``s. Then, the ``__setitem__`` would write the local `chunk` into the
            ``np.ndarray``.

            The latter allows the netcdf4 library to parallelize the write-operation by directly using the
            `netcdf4.Variable.__setitem__` method.

            Parameters
            ----------
            var : nc.Variable
                Variable to which data is to be saved.
            var_slices : tuple[int, slice]
                Keys to pass to the set-operator.
            data : DNDarray
                Data to be saved.
            data_slices: tuple[int, slice]
                As returned by the data.comm.chunk method.
            """
            slices = data_slices
            if slices is None:
                _, _, slices = data.comm.chunk(data.gshape, data.split if is_split else 0)
            start, count, stride, _ = nc.utils._StartCountStride(
                elem=var_slices,
                shape=var.shape,
                dimensions=var.dimensions,
                grp=var.group(),
                datashape=data.shape,
                put=True,
            )
            out_shape = nc._netCDF4._out_array_shape(count)
            out_split = __get_expanded_split(data.shape, out_shape, data.split)

            start, count, stride = start.T, count.T, stride.T  # transpose for iteration
            stop = start + stride * count
            new_slices = []
            for begin, end, step in zip(start, stop, stride):
                if begin.size == 1:
                    begin, end, step = begin.item(), end.item(), step.item()
                    new_slices.append(slice(begin, end, step))
                else:
                    begin, end, step = begin.flatten(), end.flatten(), step.flatten()
                    new_slices.append(
                        np.r_[
                            tuple(
                                slice(b.item(), e.item(), s.item())
                                for b, e, s in zip(begin, end, step)
                            )
                        ]
                    )
            if out_split is not None:  # add split-slice
                if isinstance(new_slices[out_split], slice):
                    start, stop, step = (
                        new_slices[out_split].start,
                        new_slices[out_split].stop,
                        new_slices[out_split].step,
                    )
                    sliced = range(start, stop, step)[slices[data.split]]
                    a, b, c = sliced.start, sliced.stop, sliced.step
                    a = None if a < 0 else a
                    b = None if b < 0 else b
                    new_slices[out_split] = slice(a, b, c)
                    # new_slices[out_split] = sliced
                elif isinstance(new_slices[out_split], np.ndarray):
                    new_slices[out_split] = new_slices[out_split][slices[data.split]]
                else:
                    new_slices[out_split] = np.r_[new_slices[out_split]][slices[data.split]]
            return tuple(new_slices)

        # attempt to perform parallel I/O if possible
        if __nc_has_par:
            try:
                with nc.Dataset(path, mode, parallel=True, comm=data.comm.handle) as handle:
                    if variable in handle.variables:
                        var = handle.variables[variable]
                    else:
                        for name, elements in zip(dimension_names, data.shape):
                            if name not in handle.dimensions:
                                handle.createDimension(name, elements if not is_unlimited else None)
                        var = handle.createVariable(
                            variable, data.dtype.char(), dimension_names, **kwargs
                        )
                    merged_slices = __merge_slices(var, file_slices, data)
                    try:
                        var[merged_slices] = (
                            data.larray.cpu() if is_split else data.larray[slices].cpu()
                        )
                    except RuntimeError:
                        var.set_collective(True)
                        var[merged_slices] = (
                            data.larray.cpu() if is_split else data.larray[slices].cpu()
                        )
            except Exception as e:
                failed = data.comm.rank + 1
                excep = e
        # otherwise a single rank only write is performed in case of local data (i.e. no split)
        elif data.comm.rank == 0:
            try:
                with nc.Dataset(path, mode) as handle:
                    if variable in handle.variables:
                        var = handle.variables[variable]
                    else:
                        for name, elements in zip(dimension_names, data.shape):
                            if name not in handle.dimensions:
                                handle.createDimension(name, elements if not is_unlimited else None)
                        var = handle.createVariable(
                            variable, data.dtype.char(), dimension_names, **kwargs
                        )
                    var.set_collective(False)  # not possible with non-parallel netcdf
                    if is_split:
                        merged_slices = __merge_slices(var, file_slices, data)
                        var[merged_slices] = data.larray.cpu()
                    else:
                        var[file_slices] = data.larray.cpu()
            except Exception as e:
                failed = 1
                excep = e
            finally:
                if data.comm.size > 1:
                    data.comm.isend(failed, dest=1)
                    data.comm.recv()

        # non-root
        else:
            # wait for the previous rank to finish writing its chunk, then write own part
            failed = data.comm.recv()
            try:
                # no MPI, but data is split, we have to serialize the writes
                if not failed and is_split:
                    with nc.Dataset(path, "r+") as handle:
                        var = handle.variables[variable]
                        var.set_collective(False)  # not possible with non-parallel netcdf
                        merged_slices = __merge_slices(var, file_slices, data)
                        var[merged_slices] = data.larray.cpu()
            except Exception as e:
                failed = data.comm.rank + 1
                excep = e
            finally:
                # ping the next node in the communicator, wrap around to 0 to complete barrier behavior
                next_rank = (data.comm.rank + 1) % data.comm.size
                data.comm.isend(failed, dest=next_rank)

        failed = data.comm.allreduce(failed, op=MPI.MAX)
        if failed - 1 == data.comm.rank:
            data.comm.bcast(excep, root=failed - 1)
            raise excep
        elif failed:
            excep = data.comm.bcast(excep, root=failed - 1)
            excep.args = "raised by process rank {}".format(failed - 1), *excep.args
            raise excep from None  # raise the same error but without traceback
            # because that is on a different process

    DNDarray.save_netcdf = lambda self, path, variable, mode="w", **kwargs: save_netcdf(
        self, path, variable, mode, **kwargs
    )
    DNDarray.save_netcdf.__doc__ = save_netcdf.__doc__


def load(
    path: str, *args: Optional[List[object]], **kwargs: Optional[Dict[str, object]]
) -> DNDarray:
    """
    Attempts to load data from a file stored on disk. Attempts to auto-detect the file format by determining the
    extension. Supports at least CSV files, HDF5 and netCDF4 are additionally possible if the corresponding libraries
    are installed.

    Parameters
    ----------
    path : str
        Path to the file to be read.
    args : list, optional
        Additional options passed to the particular functions.
    kwargs : dict, optional
        Additional options passed to the particular functions.

    Raises
    -------
    ValueError
        If the file extension is not understood or known.
    RuntimeError
        If the optional dependency for a file extension is not available.

    Examples
    --------
    >>> ht.load('data.h5', dataset='DATA')
    DNDarray([ 1.0000,  2.7183,  7.3891, 20.0855, 54.5981], dtype=ht.float32, device=cpu:0, split=None)
    >>> ht.load('data.nc', variable='DATA')
    DNDarray([ 1.0000,  2.7183,  7.3891, 20.0855, 54.5981], dtype=ht.float32, device=cpu:0, split=None)
    """
    if not isinstance(path, str):
        raise TypeError(f"Expected path to be str, but was {type(path)}")
    extension = os.path.splitext(path)[-1].strip().lower()

    if extension in __CSV_EXTENSION:
        return load_csv(path, *args, **kwargs)
    elif extension in __HDF5_EXTENSIONS:
        if supports_hdf5():
            return load_hdf5(path, *args, **kwargs)
        else:
            raise RuntimeError(f"hdf5 is required for file extension {extension}")
    elif extension in __NETCDF_EXTENSIONS:
        if supports_netcdf():
            return load_netcdf(path, *args, **kwargs)
        else:
            raise RuntimeError(f"netcdf is required for file extension {extension}")
    else:
        raise ValueError(f"Unsupported file extension {extension}")


def load_csv(
    path: str,
    header_lines: int = 0,
    sep: str = ",",
    dtype: datatype = types.float32,
    encoding: str = "utf-8",
    split: Optional[int] = None,
    device: Optional[str] = None,
    comm: Optional[Communication] = None,
) -> DNDarray:
    """
    Loads data from a CSV file. The data will be distributed along the axis 0.

    Parameters
    ----------
    path : str
        Path to the CSV file to be read.
    header_lines : int, optional
        The number of columns at the beginning of the file that should not be considered as data.
    sep : str, optional
        The single ``char`` or ``str`` that separates the values in each row.
    dtype : datatype, optional
        Data type of the resulting array.
    encoding : str, optional
        The type of encoding which will be used to interpret the lines of the csv file as strings.
    split : int or None : optional
        Along which axis the resulting array should be split.
        Default is ``None`` which means each node will have the full array.
    device : str, optional
        The device id on which to place the data, defaults to globally set default device.
    comm : Communication, optional
        The communication to use for the data distribution, defaults to global default

    Raises
    -------
    TypeError
        If any of the input parameters are not of correct type.

    Examples
    --------
    >>> import heat as ht
    >>> a = ht.load_csv('data.csv')
    >>> a.shape
    [0/3] (150, 4)
    [1/3] (150, 4)
    [2/3] (150, 4)
    [3/3] (150, 4)
    >>> a.lshape
    [0/3] (38, 4)
    [1/3] (38, 4)
    [2/3] (37, 4)
    [3/3] (37, 4)
    >>> b = ht.load_csv('data.csv', header_lines=10)
    >>> b.shape
    [0/3] (140, 4)
    [1/3] (140, 4)
    [2/3] (140, 4)
    [3/3] (140, 4)
    >>> b.lshape
    [0/3] (35, 4)
    [1/3] (35, 4)
    [2/3] (35, 4)
    [3/3] (35, 4)
    """
    if not isinstance(path, str):
        raise TypeError(f"path must be str, not {type(path)}")
    if not isinstance(sep, str):
        raise TypeError(f"separator must be str, not {type(sep)}")
    if not isinstance(header_lines, int):
        raise TypeError(f"header_lines must int, not {type(header_lines)}")
    if split not in [None, 0, 1]:
        raise ValueError(f"split must be in [None, 0, 1], but is {split}")

    # infer the type and communicator for the loaded array
    dtype = types.canonical_heat_type(dtype)
    # determine the comm and device the data will be placed on
    device = devices.sanitize_device(device)
    comm = sanitize_comm(comm)

    file_size = os.stat(path).st_size
    rank = comm.rank
    size = comm.size

    if split is None:
        with open(path) as f:
            data = f.readlines()
            data = data[header_lines:]
            result = []
            for line in data:
                values = line.replace("\n", "").replace("\r", "").split(sep)
                values = [float(val) for val in values]
                result.append(values)
            resulting_tensor = factories.array(
                result, dtype=dtype, split=split, device=device, comm=comm
            )

    elif split == 0:
        counts, displs, _ = comm.counts_displs_shape((file_size, 1), 0)
        # in case lines are terminated with '\r\n' we need to skip 2 bytes later
        lineter_len = 1
        # Read a chunk of bytes and count the linebreaks
        with open(path, "rb") as f:
            f.seek(displs[rank], 0)
            line_starts = []
            r = f.read(counts[rank])
            for pos, l in enumerate(r):
                if chr(l) == "\n":
                    # Check if it is part of '\r\n'
                    if chr(r[pos - 1]) != "\r":
                        line_starts.append(pos + 1)
                elif chr(l) == "\r":
                    # check if file line is terminated by '\r\n'
                    if pos + 1 < len(r) and chr(r[pos + 1]) == "\n":
                        line_starts.append(pos + 2)
                        lineter_len = 2
                    else:
                        line_starts.append(pos + 1)

            if rank == 0:
                line_starts = [0] + line_starts

            # Find the correct starting point
            total_lines = torch.empty(size, dtype=torch.int32)
            comm.Allgather(torch.tensor([len(line_starts)], dtype=torch.int32), total_lines)

            cumsum = total_lines.cumsum(dim=0).tolist()
            start = next(i for i in range(size) if cumsum[i] > header_lines)
            if rank < start:
                line_starts = []
            if rank == start:
                rem = header_lines - (0 if start == 0 else cumsum[start - 1])
                line_starts = line_starts[rem:]

            # Determine the number of columns that each line consists of
            if len(line_starts) > 1:
                columns = 1
                for li in r[line_starts[0] : line_starts[1]]:
                    if chr(li) == sep:
                        columns += 1
            else:
                columns = 0

            columns = torch.tensor([columns], dtype=torch.int32)
            comm.Allreduce(MPI.IN_PLACE, columns, MPI.MAX)

            # Share how far the processes need to reed in their last line
            last_line = file_size
            if size - start > 1:
                if rank == start:
                    last_line = torch.empty(1, dtype=torch.int32)
                    comm.Recv(last_line, source=rank + 1)
                    last_line = last_line.item()
                elif rank == size - 1:
                    first_line = torch.tensor(displs[rank] + line_starts[0] - 1, dtype=torch.int32)
                    comm.Send(first_line, dest=rank - 1)
                elif start < rank < size - 1:
                    last_line = torch.empty(1, dtype=torch.int32)
                    first_line = torch.tensor(displs[rank] + line_starts[0] - 1, dtype=torch.int32)
                    comm.Send(first_line, dest=rank - 1)
                    comm.Recv(last_line, source=rank + 1)
                    last_line = last_line.item()

            # Create empty tensor and iteratively fill it with the values
            local_shape = (len(line_starts), columns)
            actual_length = 0
            local_tensor = torch.empty(
                local_shape, dtype=dtype.torch_type(), device=device.torch_device
            )

            for ind, start in enumerate(line_starts):
                if ind == len(line_starts) - 1:
                    f.seek(displs[rank] + start, 0)
                    line = f.read(last_line - displs[rank] - start)
                else:
                    line = r[start : line_starts[ind + 1] - lineter_len]
                # Decode byte array
                line = line.decode(encoding)
                if len(line) > 0:
                    sep_values = [float(val) for val in line.split(sep)]
                    local_tensor[actual_length] = torch.tensor(sep_values, dtype=dtype.torch_type())
                    actual_length += 1

        # In case there are some empty lines in the csv file
        local_tensor = local_tensor[:actual_length]

        resulting_tensor = factories.array(
            local_tensor, dtype=dtype, is_split=0, device=device, comm=comm
        )
        resulting_tensor.balance_()

    elif split == 1:
        data = []

        with open(path) as f:
            for i in range(header_lines):
                f.readline()
            line = f.readline()
            values = line.replace("\n", "").replace("\r", "").split(sep)
            values = [float(val) for val in values]
            rows = len(values)

            chunk, displs, _ = comm.counts_displs_shape((1, rows), 1)
            data.append(values[displs[rank] : displs[rank] + chunk[rank]])
            # Read file line by line till EOF reached
            for line in iter(f.readline, ""):
                values = line.replace("\n", "").replace("\r", "").split(sep)
                values = [float(val) for val in values]
                data.append(values[displs[rank] : displs[rank] + chunk[rank]])
        resulting_tensor = factories.array(data, dtype=dtype, is_split=1, device=device, comm=comm)

    return resulting_tensor


def save_csv(
    data: DNDarray,
    path: str,
    header_lines: Iterable[str] = None,
    sep: str = ",",
    decimals: int = -1,
    encoding: str = "utf-8",
    comm: Optional[Communication] = None,
    truncate: bool = True,
):
    """
    Saves data to CSV files. Only 2D data, all split axes.

    Parameters
    ----------
    data : DNDarray
        The DNDarray to be saved to CSV.
    path : str
        The path as a string.
    header_lines : Iterable[str]
        Optional iterable of str to prepend at the beginning of the file. No
        pound sign or any other comment marker will be inserted.
    sep : str
        The separator character used in this CSV.
    decimals: int
        Number of digits after decimal point.
    encoding : str
        The encoding to be used in this CSV.
    comm : Optional[Communication]
        An optional object of type Communication to be used.
    truncate : bool
        Whether to truncate an existing file before writing, i.e. fully overwrite it.
        The sane default is True. Setting it to False will not shorten files if
        needed and thus may leave garbage at the end of existing files.
    """
    if not isinstance(path, str):
        raise TypeError(f"path must be str, not {type(path)}")
    if not isinstance(sep, str):
        raise TypeError(f"separator must be str, not {type(sep)}")
    # check this to allow None
    if not isinstance(header_lines, Iterable) and header_lines is not None:
        raise TypeError(f"header_lines must Iterable[str], not {type(header_lines)}")
    if data.split not in [None, 0, 1]:
        raise ValueError(f"split must be in [None, 0, 1], but is {data.split}")

    if os.path.exists(path) and truncate:
        if data.comm.rank == 0:
            os.truncate(path, 0)
        # avoid truncating and writing at the same time
        data.comm.handle.Barrier()

    amode = MPI.MODE_WRONLY | MPI.MODE_CREATE
    csv_out = MPI.File.Open(data.comm.handle, path, amode)

    # will be needed as an additional offset later
    hl_displacement = 0
    if header_lines is not None:
        hl_displacement = sum(len(hl) for hl in header_lines)
        # count additions everywhere, but write only on rank 0, avoiding reduce op to share final hl_displacement
        for hl in header_lines:
            if not hl.endswith("\n"):
                hl = hl + "\n"
                hl_displacement = hl_displacement + 1
            if data.comm.rank == 0 and header_lines:
                csv_out.Write(hl.encode(encoding))

    # formatting and element width
    data_min = smin(data).item()  # at least min is used twice, so cache it here
    data_max = smax(data).item()
    sign = 1 if data_min < 0 else 0
    if abs(data_max) > 0 or abs(data_min) > 0:
        pre_point_digits = int(log10(max(abs(data_max), abs(data_min)))) + 1
    else:
        pre_point_digits = 1

    dec_sep = 1
    fmt = ""
    if types.issubdtype(data.dtype, types.integer):
        decimals = 0
        dec_sep = 0
        if sign == 1:
            fmt = "%%%-dd" % (pre_point_digits + 1)
        else:
            fmt = "%%%dd" % (pre_point_digits)
    elif types.issubdtype(data.dtype, types.floating):
        if decimals == -1:
            decimals = 7 if data.dtype is types.float32 else 15
        if sign == 1:
            fmt = "%%%-d.%df" % (pre_point_digits + decimals + 2, decimals)
        else:
            fmt = "%%%d.%df" % (pre_point_digits + decimals + 1, decimals)

    # sign + decimal separator + pre separator digits + decimals (post separator)
    item_size = decimals + dec_sep + sign + pre_point_digits
    # each item is one position larger than its representation, either b/c of separator or line break
    row_width = item_size + 1
    if len(data.shape) > 1:
        row_width = data.shape[1] * (item_size + 1)

    offset = hl_displacement  # all splits
    if data.split == 0:
        _, displs = data.counts_displs()
        offset = offset + displs[data.comm.rank] * row_width
    elif data.split == 1:
        _, displs = data.counts_displs()
        offset = offset + displs[data.comm.rank] * (item_size + 1)

    for i in range(data.lshape[0]):
        # if lshape is of the form (x,), then there will only be a single element per row
        if len(data.lshape) == 1:
            row = fmt % (data.larray[i])
        else:
            if data.lshape[1] == 0:
                break
            row = sep.join(fmt % (item) for item in data.larray[i])

        if (
            data.split is None
            or data.split == 0
            or displs[data.comm.rank] + data.lshape[1] == data.shape[1]
        ):
            row = row + "\n"
        else:
            row = row + sep

        if data.split is not None or data.comm.rank == 0:
            csv_out.Write_at(offset, row.encode("utf-8"))

        offset = offset + row_width

    csv_out.Close()
    data.comm.handle.Barrier()


def save(
    data: DNDarray, path: str, *args: Optional[List[object]], **kwargs: Optional[Dict[str, object]]
):
    """
    Attempts to save data from a :class:`~heat.core.dndarray.DNDarray` to disk. An auto-detection based on the file
    format extension is performed.

    Parameters
    ----------
    data : DNDarray
        The array holding the data to be stored
    path : str
        Path to the file to be stored.
    args : list, optional
        Additional options passed to the particular functions.
    kwargs : dict, optional
        Additional options passed to the particular functions.

    Raises
    -------
    ValueError
        If the file extension is not understood or known.
    RuntimeError
        If the optional dependency for a file extension is not available.

    Examples
    --------
    >>> x = ht.arange(100, split=0)
    >>> ht.save(x, 'data.h5', 'DATA', mode='a')
    """
    if not isinstance(path, str):
        raise TypeError(f"Expected path to be str, but was {type(path)}")
    extension = os.path.splitext(path)[-1].strip().lower()

    if extension in __HDF5_EXTENSIONS:
        if supports_hdf5():
            save_hdf5(data, path, *args, **kwargs)
        else:
            raise RuntimeError(f"hdf5 is required for file extension {extension}")
    elif extension in __NETCDF_EXTENSIONS:
        if supports_netcdf():
            save_netcdf(data, path, *args, **kwargs)
        else:
            raise RuntimeError(f"netcdf is required for file extension {extension}")
    elif extension in __CSV_EXTENSION:
        save_csv(data, path, *args, **kwargs)
    else:
        raise ValueError(f"Unsupported file extension {extension}")


DNDarray.save = lambda self, path, *args, **kwargs: save(self, path, *args, **kwargs)
DNDarray.save.__doc__ = save.__doc__



"""
This file is for the general data parallel neural network classes.
"""
import warnings
import torch
import torch.distributed
import torch.nn as tnn

from collections import OrderedDict
from typing import Any, Callable, Dict, List, Union, Tuple

from .. import optim
from ..core.communication import MPI
from ..core.communication import MPI_WORLD
from ..core.communication import MPICommunication


__all__ = ["DataParallel", "DataParallelMultiGPU"]


class DataParallel(tnn.Module):
    """
    Implements data parallelism across multiple processes. This means that the same model will be run locally
    on each process. Creation of the model is similar to PyTorch, the only changes are using HeAT layers (ht.nn.layer)
    in the initialization of the network/optimizer. If there is not a HeAT layer, it will fall back to the PyTorch layer
    of the same name. The same is true for the optimizer. It's possible to use more than one optimizer, but
    communication during parameter updates is limited to blocking. The same limitation takes effect when passing an
    optimizer that does not deal exactly with the set of model's parameters. For the given model both the
    ``__init__()`` and ``forward()`` functions must be defined in the class defining the network.

    An example of this is shown in `examples/mnist.py <https://github.com/helmholtz-analytics/heat/blob/504-docstring-formatting/examples/nn/mnist.py>`_.

    It is highly recommended that a HeAT DataLoader is used, see :func:`ht.utils.data.DataLoader <heat.utils.data.datatools.DataLoader>`.
    The default communications scheme for this is blocking. The blocking scheme will average the model parameters during
    the backwards step, synchronizing them before the next model iteration.

    Usage of more than one optimizer forces MPI communication to be parameter updates to use blocking communications.

    Attributes
    ----------
    module : torch.nn.Module
        The local module
    comm : MPICommunication
        Communicator to use
    optimizer : heat.DataParallelOptimizer, List, Tuple
        Individual or sequence of DataParallelOptimizers to be used
    blocking_parameter_updates : bool, optional
        Flag indicating the usage of blocking communications for parameter updates
        Default: non-blocking updates (``False``)
    """

    def __init__(
        self,
        module: torch.nn.Module,
        comm: MPICommunication,
        optimizer: Union[optim.DataParallelOptimizer, List, Tuple],
        blocking_parameter_updates: bool = False,
    ):  # noqa: D107
        if isinstance(optimizer, optim.DASO):
            raise TypeError(
                "For use with DASO please use DataParallelMultiGPU instead of DataParallel"
            )
        super(DataParallel, self).__init__()
        self.module = module
        self.comm = comm
        self.blocking_parameter_updates = blocking_parameter_updates

        self._dp_optimizers = []
        self._layer_wait_handles = OrderedDict()
        self._fwd_hook_handles = []
        # set of layers' names with active wait handles (only relevant for non-blocking)
        self._active_layers = set()
        # slices of parameters belonging to one and the same layer
        self._param_slices = {}
        # pytorch internal parameter indexing
        self._param_indices = {}

        # raise error if no DP optimizer is given
        if not isinstance(optimizer, (list, tuple)):
            optimizer = [optimizer]
        for i in optimizer:
            if not isinstance(i, optim.DataParallelOptimizer):
                raise TypeError("optimizers must be optim.DataParallelOptimizer")

        # current implementation of non-blocking communication during parameter updates has some limitations that cause
        # fallback onto blocking in case of overstepping them
        if not self.blocking_parameter_updates and (
            len(optimizer) > 1
            or list(module.parameters()) != optimizer[0].torch_optimizer.param_groups[0]["params"]
        ):
            self.blocking_parameter_updates = True
            warnings.warn(
                "Usage of more than one DataParallelOptimizer causes fallback on blocking MPI "
                "communication during parameter updates.",
                stacklevel=2,
            )

        # assign given optimizers to this model
        for dp_optimizer in optimizer:
            self._dp_optimizers.append(dp_optimizer)
            dp_optimizer.blocking_parameter_updates = self.blocking_parameter_updates

        # unify parameters across nodes by unifying the random seed and resetting parameters
        torch.random.manual_seed(2147483646)  # max int32 value - 1
        self.module.apply(self._reset_parameters)

        # get parameter indexing and slices
        start_idx = 0
        layer_name_prev = None
        for idx, (name, param) in enumerate(module.named_parameters()):
            self._param_indices[name] = idx
            layer_name = name.rsplit(sep=".", maxsplit=1)[0]
            if layer_name_prev is None:
                layer_name_prev = layer_name
            if layer_name_prev != layer_name:
                self._param_slices[layer_name_prev] = slice(start_idx, idx)
                layer_name_prev = layer_name
                start_idx = idx

            # register backward hooks for all model parameter tensors
            if self.blocking_parameter_updates:
                param.register_hook(self._blocking_hook)
            else:
                param.register_hook(self._nonblocking_hook(layer_name, name))
        self._param_slices[layer_name_prev] = slice(start_idx, len(self._param_indices))

    def __setattr__(self, name: str, value: Union[torch.nn.Module, torch.Tensor, Any]) -> None:
        """
        Overwrite the current torch.nn.Module.__setattr__ so that it auto-detects the end of epoch's
        training phase and finalize wait handles (only relevant for non-blocking)
        """
        if name == "training" and not value and not self.blocking_parameter_updates:
            self._iparam_update()
        super(DataParallel, self).__setattr__(name, value)

    def forward(self, *inputs: tuple, **kwargs: dict) -> torch.Tensor:
        """
        Do the forward step for the network, receive the parameters from the last
        """
        # check if non-blocking and training
        if not self.blocking_parameter_updates and self.module.training:
            # register forward hooks for all layers with wait handles
            for name, submodule in self.module.named_modules():
                if name == "":
                    continue
                if name in self._layer_wait_handles:
                    hook_handle = submodule.register_forward_pre_hook(self._forward_hook(name))
                    self._fwd_hook_handles.append(hook_handle)

        # perform forward pass
        ret = self.module(*inputs, **kwargs)
        # finalize potentially remaining wait handles and update corresponding params (if
        # computation graph has changed between previous backward and this forward)
        if not self.blocking_parameter_updates and self.module.training:
            # set has to be copied in order to be manipulated during iteration
            active_layers_cpy = self._active_layers.copy()
            for layer_name in active_layers_cpy:
                self._forward_hook(layer_name)(None, None)
        # reset optimizer flag
        for ldp_optimizer in self._dp_optimizers:
            ldp_optimizer.update_next = False
        # clear dictionary after all wait handles are used up (dynamic computation graph)
        self._layer_wait_handles.clear()
        # remove forward hooks (dynamic computation graph)
        for hook_handle in self._fwd_hook_handles:
            hook_handle.remove()
        self._fwd_hook_handles.clear()

        return ret

    def _iparam_update(self, param_slice: slice = None, layer_names: List[str] = None) -> None:
        r"""
        Update parameters asynchronously via wait handles.

        Parameters
        ----------
        param_slice : slice, optional
            Slice object for creating a view onto optimizer's params list.\n
            By default, the whole params list is used, (``None``)
        layer_names : list(str), optional
            List of layer names which parameters will be updated, must match param_slice.\n
            By default, all layers are updated (``None``)
        """
        # for non-blocking, only one dp optimizer is allowed
        dp_optimizer = self._dp_optimizers[0]
        # perform update on the whole model
        if param_slice is None:
            param_slice = slice(len(dp_optimizer.params_ref))
        if layer_names is None:
            layer_names = list(self._layer_wait_handles.keys())
        # update params that are visible for the optimizer
        dp_optimizer.torch_optimizer.param_groups[0]["params"] = dp_optimizer.params_ref[
            param_slice
        ]
        # iterate over layers
        for layer_name in reversed(layer_names):
            # only perform update, if all given layers hold unfinalized wait handles (important for layer reuse)
            if layer_name not in self._active_layers:
                return
            # iterate over layer's parameters/associated wait handles
            for param_name, wait_handle, dtp, tens in self._layer_wait_handles[layer_name]:
                # get internal index of selected parameter
                param_idx = self._param_indices[param_name]
                # synchronize, get parameter's global gradient
                wait_handle.wait()
                # check if shapes are matching
                if (
                    dp_optimizer.params_ref[param_idx].grad.data.shape != tens.shape
                ):  # wait_handle.tensor.shape:
                    raise ValueError("Shapes must be equal.")
                # accumulate parameter's global gradient
                dp_optimizer.params_ref[param_idx].grad.data += tens.to(dtp)  # wait_handle.tensor
                # remove layer from set of active layers, if present
                self._active_layers.discard(layer_name)
        # if desired, perform actual parameter update
        if dp_optimizer.update_next:
            dp_optimizer.torch_optimizer.step()

    def _blocking_hook(self, grad_loc: torch.Tensor) -> torch.Tensor:
        """
        Add a blocking hook to the PyTorch DAG for all of the backwards calls.

        Parameters
        ----------
        grad_loc : torch.Tensor
            The local gradient

        References
        ----------
        [1] (cf. https://pytorch.org/docs/stable/tensors.html#torch.Tensor.register_hook).
        """
        grad_loc_bf = grad_loc.to(torch.float)  # bfloat16)
        # average local gradients
        grad_loc_bf *= 1 / float(self.comm.size)
        # perform MPI Allreduce to compute global gradient
        self.comm.Allreduce(MPI.IN_PLACE, grad_loc_bf, MPI.SUM)  # mpi_sum_bf16)
        return grad_loc_bf.to(grad_loc.dtype)

    def _nonblocking_hook(self, layer_name: str, param_name: str) -> Callable:
        """
        Add a nonblocking hook to send and receive the averaged parameters after the backwards step

        Parameters
        ----------
        layer_name : str
            Name of the layer
        param_name : str
            Name of the parameter
        """
        # hook function for blocking gradient data exchange

        def _hook(grad_loc: torch.Tensor) -> torch.Tensor:
            with torch.no_grad():
                wrk = grad_loc.to(torch.float)  # bfloat16)
            # counterbalance local gradient averaging
            wrk *= 1 / float(self.comm.size)
            # perform MPI IAllreduce to compute global gradient, returns wait handle
            wait_handle = self.comm.Iallreduce(MPI.IN_PLACE, wrk, MPI.SUM)  # mpi_sum_bf16)
            # if layer wait handle dict does not contain the layer, add it -> automatically tracks reversed layer order
            if layer_name not in self._layer_wait_handles:
                self._layer_wait_handles[layer_name] = []
            # add layer to set of active layers
            self._active_layers.add(layer_name)
            # assign wait handle to its layer, layer-internal sorting by size
            # bisect.insort(
            #     self._layer_wait_handles[layer_name], (wrk.numel(), param_name, wait_handle)
            # )
            # TODO: is sorting faster? or is there any difference?
            self._layer_wait_handles[layer_name].append(
                (param_name, wait_handle, grad_loc.dtype, wrk)
            )
            # don't return grad_loc, otherwise gradient is doubled
            return torch.zeros(*wrk.size(), device=grad_loc.device)

        return _hook

    def _forward_hook(self, layer_name: str) -> Callable:
        """
        Add a forward hook to update parameters during the forward step. This will return a hook with can be added
        using the ``submodule.register_forward_pre_hook`` command.

        Parameters
        ----------
        layer_name : str
            Name of the layer
        """
        # hook function for non-blocking parameter update

        def _hook(_, input_):
            # update parameters of given layer
            param_slice = self._param_slices[layer_name]
            self._iparam_update(param_slice, [layer_name])
            return input_

        return _hook

    @staticmethod
    def _reset_parameters(module: tnn.Module) -> None:
        """
        Reset parameters of given torch submodule. Only works for basic module types containing ``reset_parameters``
        function.

        Parameters
        ----------
        module: torch.nn.Module
            Submodule whose parameters are to be reset
        """
        if callable(getattr(module, "reset_parameters", None)):
            module.reset_parameters()


class DataParallelMultiGPU(tnn.Module):
    """
    This creates data parallel networks local to each node using PyTorch's distributed class. This does NOT
    do any global synchronizations. To make optimal use of this structure, use :func:`ht.optim.DASO <heat.optim.dp_optimizer.DASO>`.

    Notes
    -----
    The PyTorch distributed process group must already exist before this class is initialized.

    Parameters
    ----------
    module: torch.nn.Module
        an implemented PyTorch model
    optimizer: optim.DASO
        A DASO optimizer. Other optimizers are not yet implemented. The DASO optimizer should be
        defined prior to calling this class.
    comm: MPICommunication, optional
        A global communicator.
        Default: :func:`MPICommunication <heat.core.comm.MPICommunication>`
    """

    def __init__(
        self, module: torch.nn.Module, optimizer: optim.DASO, comm: MPICommunication = MPI_WORLD
    ):  # noqa: D107
        super(DataParallelMultiGPU, self).__init__()
        rank = comm.rank
        if torch.cuda.device_count() > 1:
            self.loc_gpus = torch.cuda.device_count()
            local_rank = rank % self.loc_gpus
            device = f"cuda:{str(local_rank)}"
            torch.cuda.set_device(device=device)
            module = tnn.parallel.DistributedDataParallel(module, device_ids=[local_rank])
        else:
            warnings.warn(
                "DataParallelMultiGPU should be used with multiple GPUs per node", UserWarning
            )
        self.module = module
        self.comm = comm

        # unify parameters across nodes by unifying the random seed and resetting parameters
        self.module.apply(self._reset_parameters)

        optimizer.set_model(self.module)

    def forward(self, *inputs: Tuple, **kwargs: Dict) -> torch.Tensor:
        """
        Calls the forward method for the torch model
        """
        return self.module(*inputs, **kwargs)

    @staticmethod
    def _reset_parameters(module: tnn.Module) -> None:
        """
        Reset parameters of given torch submodule. Only works for basic module types containing ``reset_parameters``
        function.

        Parameters
        ----------
        module: torch.nn.Module
            Submodule whose parameters are to be reset
        """
        if callable(getattr(module, "reset_parameters", None)):
            module.reset_parameters()



"""Create a sperical dataset."""
import heat as ht


def create_spherical_dataset(
    num_samples_cluster, radius=1.0, offset=4.0, dtype=ht.float32, random_state=1
):
    """
    Creates k=4 sperical clusters in 3D space along the space-diagonal

    Parameters
    ----------
    num_samples_cluster: int
        Number of samples per cluster. Each process will create n // MPI_WORLD.size elements for each cluster
    radius: float
        Radius of the sphere
    offset: float
        Shift of the clusters along the axes. The 4 clusters will be positioned centered around c1=(offset, offset,offset),
        c2=(2*offset,2*offset,2*offset), c3=(-offset, -offset, -offset) and c4=(2*offset, -2*offset, -2*offset)
    dtype: ht.datatype
        Dataset dtype
    random_state: int
        seed of the torch random number generator
    """
    # contains num_samples

    p = ht.MPI_WORLD.size
    # create k sperical clusters with each n elements per cluster. Each process creates k * n/p elements
    num_ele = num_samples_cluster // p
    ht.random.seed(random_state)
    # radius between 0 and 1
    r = ht.random.rand(num_ele, split=0) * radius
    # theta between 0 and pi
    theta = ht.random.rand(num_ele, split=0) * 3.1415
    # phi between 0 and 2pi
    phi = ht.random.rand(num_ele, split=0) * 2 * 3.1415
    # Cartesian coordinates
    x = r * ht.sin(theta) * ht.cos(phi)
    x.astype(dtype, copy=False)
    y = r * ht.sin(theta) * ht.sin(phi)
    y.astype(dtype, copy=False)
    z = r * ht.cos(theta)
    z.astype(dtype, copy=False)

    cluster1 = ht.stack((x + offset, y + offset, z + offset), axis=1)
    cluster2 = ht.stack((x + 2 * offset, y + 2 * offset, z + 2 * offset), axis=1)
    cluster3 = ht.stack((x - offset, y - offset, z - offset), axis=1)
    cluster4 = ht.stack((x - 2 * offset, y - 2 * offset, z - 2 * offset), axis=1)

    data = ht.concatenate((cluster1, cluster2, cluster3, cluster4), axis=0)
    # Note: enhance when shuffel is available
    return data



"""
Function and classes useful for loading data into neural networks
"""

import torch
from torch.utils import data as torch_data
from typing import Callable, List, Iterator, Union, Optional, Sized

from ...core.dndarray import DNDarray
from ...core.communication import MPI_WORLD
from . import partial_dataset

__all__ = ["DataLoader", "Dataset", "dataset_shuffle", "dataset_ishuffle"]


class DataLoader:
    r"""
    The combines either a :func:`DNDarray <heat.core.dndarray.DNDarray>` or a torch `Dataset <https://pytorch.org/docs/stable/data.html?highlight=dataset#torch.utils.data.Dataset>`_
    with a sampler. This provides an iterable over the local dataset and it will shuffle the data at the end of the
    iterator. If a :func:`DNDarray <heat.core.dndarray.DNDarray>` is given, then a :func:`Dataset` will be created
    internally.

    Currently, this only supports only map-style datasets with single-process loading. It uses the random
    batch sampler. The rest of the ``DataLoader`` functionality mentioned in `torch.utils.data.dataloader <https://pytorch.org/docs/stable/data.html?highlight=dataset#torch.utils.data.DataLoader>`_ applies.

    Arguments:
        dataset : :func:`Dataset`, torch `Dataset <https://pytorch.org/docs/stable/data.html?highlight=dataset#torch.utils.data.Dataset>`_, :func:`heat.utils.data.partial_dataset.PartialH5Dataset`
            A torch dataset from which the data will be returned by the created iterator
        batch_size : int, optional
            How many samples per batch to load\n
             Default: 1
        num_workers : int, optional
            How many subprocesses to use for data loading. 0 means that the data will be loaded in the main process.\n
            Default: 0
        collate_fn : callable, optional
            Merges a list of samples to form a mini-batch of torch.Tensor(s).  Used when using batched loading from a
            map-style dataset.\n
            Default: None
        pin_memory : bool, optional
            If ``True``, the data loader will copy torch.Tensors into CUDA pinned memory before returning them.
            If your data elements are a custom type, or your :attr:`collate_fn` returns a batch that is a custom type,
            see the example below. \n
            Default: False
        drop_last : bool, optional
            Set to ``True`` to drop the last incomplete batch, if the dataset size is not divisible by
            the batch size. If ``False`` and the size of dataset is not divisible by the batch size, then
            the last batch will be smaller.\n
            Default: ``False``
        timeout : int or float, optional
            If positive, the timeout value for collecting a batch from workers. Should always be non-negative.\n
            Default: 0
        worker_init_fn : callable, optional
            If not ``None``, this will be called on each worker subprocess with the worker id
            (an int in ``[0, num_workers - 1]``) as input, after seeding and before data loading.\n
            default: None

    Attributes
    ----------
    dataset : :func:`Dataset`, torch `Dataset <https://pytorch.org/docs/stable/data.html?highlight=dataset#torch.utils.data.Dataset>`_, :func:`heat.utils.data.partial_dataset.PartialH5Dataset`
        The dataset created from the local data
    DataLoader : `torch.utils.data.dataloader <https://pytorch.org/docs/stable/data.html?highlight=dataset#torch.utils.data.DataLoader>`_
        The local DataLoader object. Used in the creation of the iterable and the length
    _first_iter : bool
        Flag indicating if the iterator created is the first one. If it is not, then the data will be shuffled before
        the iterator is created
    last_epoch : bool
        Flag indicating last epoch
    """

    def __init__(
        self,
        dataset: Union[torch_data.Dataset, partial_dataset.PartialH5Dataset],
        batch_size: int = 1,
        num_workers: int = 0,
        collate_fn: Callable = None,
        pin_memory: bool = False,
        drop_last: bool = False,
        timeout: Union[int, float] = 0,
        worker_init_fn: Callable = None,
    ):  # noqa: D107
        if not isinstance(dataset, (torch_data.Dataset, Dataset, partial_dataset.PartialH5Dataset)):
            raise TypeError(
                f"dataset must be a torch Dataset, heat Dataset, heat PartialH5Dataset, currently: {type(dataset)}"
            )
        self.dataset = dataset
        self.ishuffle = self.dataset.ishuffle
        if isinstance(self.dataset, partial_dataset.PartialH5Dataset):
            drop_last = True

        self.DataLoader = torch_data.DataLoader(
            dataset=self.dataset,
            batch_size=batch_size,
            shuffle=True,
            batch_sampler=None,
            num_workers=num_workers,
            collate_fn=collate_fn,
            drop_last=drop_last,
            pin_memory=pin_memory,
            timeout=timeout,
            worker_init_fn=worker_init_fn,
        )
        self._first_iter = True
        self.last_epoch = False

    def __iter__(self) -> Iterator:
        """
        Generate a new iterator of a type dependent on the type of dataset.
        Returns a :class:`partial_dataset.PartialH5DataLoaderIter` if the dataset is a :class:`partial_dataset.PartialH5Dataset`
        :func:`self._full_dataset_shuffle_iter` otherwise
        """
        if isinstance(self.dataset, partial_dataset.PartialH5Dataset):
            return partial_dataset.PartialH5DataLoaderIter(self)
        if hasattr(self, "_full_dataset_shuffle_iter"):
            # if it is a normal heat dataset then this is defined
            self._full_dataset_shuffle_iter()
        return self.DataLoader.__iter__()

    def __len__(self) -> int:
        """
        Get the length of the dataloader. Returns the number of batches.
        """
        return self.DataLoader.__len__()

    def _full_dataset_shuffle_iter(self):
        # logic for when to shuffle the data
        if not self.ishuffle:
            if self._first_iter:
                self._first_iter = False
            else:
                # shuffle after the first epoch but before the iterator is generated
                self.dataset.Shuffle()
        else:
            # start the shuffling for the next iteration
            if not self.last_epoch:
                self.dataset.Ishuffle()

            if self._first_iter:
                self._first_iter = False
            else:
                dataset_irecv(self.dataset)


class Dataset(torch_data.Dataset):
    r"""
    An abstract class representing a given dataset. This inherits from torch.utils.data.Dataset.

    This class is a general example for what should be done to create a Dataset. When creating a dataset all of the
    standard attributes should be set, the ``__getitem__``, ``__len__``, and ``shuffle`` functions must be defined.

        - ``__getitem__`` : how an item is given to the network
        - ``__len__`` : the number of data elements to be given to the network in total
        - ``Shuffle()`` : how the data should be shuffled between the processes. The function shown below is for a dataset composed of only data and without targets. The function :func:`dataset_shuffle` abstracts this. For this function only the dataset and a list of attributes to shuffle are given.\n
        - ``Ishuffle()`` : A non-blocking version of ``Shuffle()``, this is handled in the abstract function :func:`dataset_ishuffle`. It works similarly to :func:`dataset_shuffle`.

    As the amount of data across processes can be non-uniform, the dataset class will slice off the remaining elements
    on whichever processes have more data than the others. This should only be 1 element.
    The shuffle function will shuffle all of the data on the process.

    It is recommended that for ``DNDarray`` s, the split is either 0 or None

    Parameters
    ----------
    array : DNDarray
        DNDarray for which to great the dataset
    transform : Callable
        Transformation to call before a data item is returned
    ishuffle : bool, optional
        flag indicating whether to use non-blocking communications for shuffling the data between epochs
        Note: if ``True``, the ``Ishuffle()`` function must be defined within the class\n
        Default: False

    Attributes
    ----------
    These are the required attributes.

    htdata : DNDarray
        Full data
    _cut_slice : slice
        Slice to cut off the last element to get a uniform amount of data on each process
    comm : MPICommunicator
        Communication object used to send the data between processes
    lcl_half : int
        Half of the number of data elements on the process
    data : torch.Tensor
        The local data to be used in training
    transforms : Callable
        Transform to be called during the getitem function
    ishuffle : bool
        Flag indicating if non-blocking communications are used for shuffling the data between epochs
    """

    def __init__(
        self,
        array,
        transforms: Optional[Union[List, Callable]] = None,
        ishuffle: Optional[bool] = False,
        test_set: Optional[bool] = False,
    ):  # noqa: D107
        self.htdata = array
        self.comm = array.comm
        self.test_set = test_set
        # create a slice to create a uniform amount of data on each process
        min_data_split = array.gshape[array.split] // array.comm.size
        self.lcl_half = min_data_split // 2
        arb_slice = [slice(None)] * array.ndim
        arb_slice[array.split] = slice(min_data_split)
        self._cut_slice = tuple(arb_slice)
        self.data = array._DNDarray__array[self._cut_slice]
        if not isinstance(transforms, (list, tuple)) and transforms is not None:
            transforms = [transforms]
        self.transforms = transforms
        self.ishuffle = ishuffle

    def __getitem__(self, index: Union[int, slice, tuple, list, torch.Tensor]) -> torch.Tensor:
        """
        This is the most basic form of getitem. As the dataset is often very specific to the dataset,
        this should be overwritten by the user. In this form it only gets the raw items from the data.
        """
        if self.transforms:
            return self.transforms[0](self.data[index])
        return self.data[index]

    def __len__(self) -> int:
        """
        Get the number of items in the dataset. This should be overwritten by custom datasets
        """
        return self.data.shape[0]

    def Shuffle(self):
        """
        Send half of the local data to the process ``self.comm.rank + 1`` if available, else wrap around. After
        receiving the new data, shuffle the local tensor.
        """
        if not self.test_set:
            dataset_shuffle(dataset=self, attrs=[["data", "htdata"]])

    def Ishuffle(self):
        """
        Send half of the local data to the process ``self.comm.rank + 1`` if available, else wrap around. After
        receiving the new data, shuffle the local tensor.
        """
        if not self.test_set:
            dataset_ishuffle(dataset=self, attrs=[["data", "htdata"]])


def dataset_shuffle(dataset: Union[Dataset, torch_data.Dataset], attrs: List[list]):
    """
    Shuffle the given attributes of a dataset across multiple processes. This will send half of the data to rank + 1.
    Once the new data is received, it will be shuffled into the existing data on the process.
    This function will be called by the DataLoader automatically if ``dataset.ishuffle = False``.
    attrs should have the form [[torch.Tensor, DNDarray], ... i.e. [['data', 'htdata`]] assume that all of the attrs have the same dim0 shape as the local data

    Parameters
    ----------
    dataset : Dataset
        the dataset to shuffle
    attrs : List[List[str, str], ... ]
        List of lists each of which contains 2 strings. The strings are the handles corresponding to the Dataset
        attributes corresponding to the global data DNDarray and the local data of that array, i.e. [["data, "htdata"],]
        would shuffle the htdata around and set the correct amount of data for the ``dataset.data`` attribute. For
        multiple parameters multiple lists are required. I.e. [["data", "htdata"], ["targets", "httargets"]]

    Notes
    -----
    ``dataset.comm`` must be defined for this function to work.
    """
    # attrs -> [[torch.Tensor, DNDarray], ...]
    if attrs[0][1] is not None:
        prm = torch.randperm(getattr(dataset, attrs[0][1])._DNDarray__array.shape[0])
    else:
        prm = torch.randperm(getattr(dataset, attrs[0][0]).shape[0])
    comm = dataset.comm
    for att in attrs:
        ld = getattr(dataset, att[0])
        snd = ld[: dataset.lcl_half].clone()
        snd_shape, snd_dtype, snd_dev = snd.shape, snd.dtype, snd.device
        dest = comm.rank + 1 if comm.rank + 1 != comm.size else 0
        # send the top half of the data to the next process
        send_wait = comm.Isend(snd, dest=dest)
        del snd
        new_data = torch.empty(snd_shape, dtype=snd_dtype, device=snd_dev)
        src = comm.rank - 1 if comm.rank != 0 else comm.size - 1
        rcv_w = comm.Irecv(new_data, source=src)
        send_wait.wait()
        rcv_w.wait()
        # set the DNDarray data
        if att[1] is not None:
            getattr(dataset, att[1])._DNDarray__array[: dataset.lcl_half] = new_data
            # shuffle all of the data around
            shuffled = getattr(dataset, att[1])._DNDarray__array[prm]
            getattr(dataset, att[1])._DNDarray__array = shuffled
            # set the torch data
            setattr(dataset, att[0], shuffled[dataset._cut_slice])
        else:
            getattr(dataset, att[0])[: dataset.lcl_half] = new_data
            # shuffle all of the data around
            shuffled = getattr(dataset, att[0])[prm]
            setattr(dataset, att[0], shuffled[dataset._cut_slice])


def dataset_ishuffle(dataset: Union[Dataset, torch_data.Dataset], attrs: List[list]):
    """
    Shuffle the given attributes of a dataset across multiple processes, using non-blocking communications.
    This will send half of the data to rank + 1. The data must be received by the :func:`dataset_irecv` function.

    This function will be called by the DataLoader automatically if ``dataset.ishuffle = True``. This is set either
    during the definition of the class of its initialization by a given paramete.

    Parameters
    ----------
    dataset : Dataset
        the dataset to shuffle
    attrs : List[List[str, str], ... ]
        List of lists each of which contains 2 strings. The strings are the handles corresponding to the Dataset
        attributes corresponding to the global data DNDarray and the local data of that array, i.e. [["htdata, "data"],]
        would shuffle the htdata around and set the correct amount of data for the ``dataset.data`` attribute. For
        multiple parameters multiple lists are required. I.e. [["htdata", "data"], ["httargets", "targets"]]

    Notes
    -----
    ``dataset.comm`` must be defined for this function to work.
    """
    # attrs should have the form [[heat array, sliced array], [...], ...]
    #       i.e. [['data', 'htdata']]
    # assume that all of the attrs have the same dim0 shape as the local data
    comm = dataset.comm
    ret_list = []
    for att in attrs:
        snd = getattr(dataset, att[0])[: dataset.lcl_half].clone()
        snd_shape, snd_dtype, snd_dev = snd.shape, snd.dtype, snd.device
        dest = comm.rank + 1 if comm.rank + 1 != comm.size else 0
        # send the top half of the data to the next process
        send_wait = comm.Isend(snd, dest=dest, tag=99999)
        new_data = torch.empty(snd_shape, dtype=snd_dtype, device=snd_dev)
        src = comm.rank - 1 if comm.rank != 0 else comm.size - 1
        wait = comm.Irecv(new_data, source=src, tag=99999)
        ret_list.append([att, wait, new_data])
        send_wait.wait()
        del snd
    setattr(dataset, "rcv_list", ret_list)


def dataset_irecv(dataset: Union[Dataset, torch_data.Dataset]):
    """
    Receive the data sent by the :func:`dataset_ishuffle` function. This will wait for the data and then shuffle the
    data into the existing data on the process

    This function will be called by the DataLoader automatically if ``dataset.ishuffle = True``. This is set either
    during the definition of the class of its initialization by a given paramete.

    Parameters
    ----------
    dataset : Dataset
        the dataset to shuffle

    Notes
    -----
    ``dataset.comm`` must be defined for this function to work.
    """
    setattr(dataset, "shuffle_prm", torch.randperm(dataset.data.shape[0]))
    rcv_list = getattr(dataset, "rcv_list")
    prm = getattr(dataset, "shuffle_prm")
    for rcv in rcv_list:
        rcv[1].wait()
        if rcv[0][1] is not None:
            getattr(dataset, rcv[0][1])._DNDarray__array[: dataset.lcl_half] = rcv[2]
            # shuffle all of the data around
            shuffled = getattr(dataset, rcv[0][1])._DNDarray__array[prm]
            getattr(dataset, rcv[0][1])._DNDarray__array = shuffled
            # set the torch data
            setattr(dataset, rcv[0][0], shuffled[dataset._cut_slice])
        else:
            getattr(dataset, rcv[0][0])[: dataset.lcl_half] = rcv[2]
            # shuffle all of the data around
            shuffled = getattr(dataset, rcv[0][0])[prm]
            setattr(dataset, rcv[0][0], shuffled[dataset._cut_slice])



"""
This module changes the internal memory of an array.
"""

import torch

from . import sanitation
from .dndarray import DNDarray

__all__ = ["copy", "sanitize_memory_layout"]


def copy(x: DNDarray) -> DNDarray:
    """
    Return a deep copy of the given object.

    Parameters
    ----------
    x : DNDarray
        Input array to be copied.

    Examples
    --------
    >>> a = ht.array([1,2,3])
    >>> b = ht.copy(a)
    >>> b
    DNDarray([1, 2, 3], dtype=ht.int64, device=cpu:0, split=None)
    >>> a[0] = 4
    >>> a
    DNDarray([4, 2, 3], dtype=ht.int64, device=cpu:0, split=None)
    >>> b
    DNDarray([1, 2, 3], dtype=ht.int64, device=cpu:0, split=None)
    """
    sanitation.sanitize_in(x)
    return DNDarray(x.larray.clone(), x.shape, x.dtype, x.split, x.device, x.comm, x.balanced)


DNDarray.copy = lambda self: copy(self)
DNDarray.copy.__doc__ = copy.__doc__


def sanitize_memory_layout(x: torch.Tensor, order: str = "C") -> torch.Tensor:
    """
    Return the given object with memory layout as defined below. The default memory distribution is assumed.

    Parameters
    -----------
    x: torch.Tensor
        Input data
    order: str, optional.
        Default is ``'C'`` as in C-like (row-major) memory layout. The array is stored first dimension first (rows first if ``ndim=2``).
        Alternative is ``'F'``, as in Fortran-like (column-major) memory layout. The array is stored last dimension first (columns first if ``ndim=2``).
    """
    if order == "K":
        raise NotImplementedError(
            "Internal usage of torch.clone() means losing original memory layout for now. \n Please specify order='C' for row-major, order='F' for column-major layout."
        )
    if x.ndim < 2 or x.numel() == 0:
        # do nothing
        return x
    stride_diff = torch.as_tensor(x.stride())
    # since strides can get a bit wonky with operations like transpose
    #   we should assume that the tensors are row major or are distributed the default way
    stride_diff = stride_diff[1:] - stride_diff[:-1]
    column_major = torch.all(stride_diff >= 0)
    row_major = not column_major
    del stride_diff
    if (order == "C" and row_major) or (order == "F" and column_major):
        # do nothing
        del column_major, row_major
        return x
    if (order == "C" and column_major) or (order == "F" and row_major):
        dims = list(range(x.ndim))
        dims = tuple(reversed(dims))
        y = torch.empty_like(x)
        permutation = x.permute(dims).contiguous()
        y = y.set_(
            permutation.storage(),
            x.storage_offset(),
            x.shape,
            tuple(reversed(permutation.stride())),
        )
        del permutation, dims, column_major, row_major, x
        return y
    else:
        raise ValueError(
            f"combination of order and layout not permitted, order: {order} column major: {column_major} row major: {row_major}"
        )



"""
Base-module for k-clustering algorithms
"""

import heat as ht
from typing import Optional, Union, Callable
from heat.core.dndarray import DNDarray


class _KCluster(ht.ClusteringMixin, ht.BaseEstimator):
    """
    Base class for k-statistics clustering algorithms (kmeans, kmedians, kmedoids).
    The clusters are represented by centroids ci (we use the term from kmeans for simplicity)

    Parameters
    ----------
    metric : function
        One of the distance metrics in ht.spatial.distance. Needs to be passed as lambda function to take only two arrays as input
    n_clusters : int
        The number of clusters to form as well as the number of centroids to generate.
    init : str or DNDarray, default: ‘random’
        Method for initialization:

        - ‘probability_based’ : selects initial cluster centers for the clustering in a smart way to speed up convergence (k-means++)
        - ‘random’: choose k observations (rows) at random from data for the initial centroids.
        - ``DNDarray``: gives the initial centers, should be of Shape = (n_clusters, n_features)
    max_iter : int
        Maximum number of iterations for a single run.
    tol : float, default: 1e-4
        Relative tolerance with regards to inertia to declare convergence.
    random_state : int
        Determines random number generation for centroid initialization.
    """

    def __init__(
        self,
        metric: Callable,
        n_clusters: int,
        init: Union[str, DNDarray],
        max_iter: int,
        tol: float,
        random_state: int,
    ):  # noqa: D107
        self.n_clusters = n_clusters
        self.init = init
        self.max_iter = max_iter
        self.tol = tol
        self.random_state = random_state

        # in-place properties
        self._metric = metric
        self._cluster_centers = None
        self._labels = None
        self._inertia = None
        self._n_iter = None

    @property
    def cluster_centers_(self) -> DNDarray:
        """
        Returns the coordinates of the cluster centers.
        If the algorithm stops before fully converging (see ``tol`` and ``max_iter``),
        these will not be consistent with :func:`labels_`.
        """
        return self._cluster_centers

    @property
    def labels_(self) -> DNDarray:
        """
        Returns the labels of each point
        """
        return self._labels

    @property
    def inertia_(self) -> float:
        """
        Returns the sum of squared distances of samples to their closest cluster center.
        """
        return self._inertia

    @property
    def n_iter_(self) -> int:
        """
        Returns the number of iterations run.
        """
        return self._n_iter

    def _initialize_cluster_centers(self, x: DNDarray):
        """
        Initializes the K-Means centroids.

        Parameters
        ----------
        x : DNDarray
            The data to initialize the clusters for. Shape = (n_samples, n_features)
        """
        # always initialize the random state
        if self.random_state is not None:
            ht.random.seed(self.random_state)

        # initialize the centroids by randomly picking some of the points
        if self.init == "random":
            # Samples will be equally distributed drawn from all involved processes
            _, displ, _ = x.comm.counts_displs_shape(shape=x.shape, axis=0)
            centroids = ht.empty(
                (self.n_clusters, x.shape[1]), split=None, device=x.device, comm=x.comm
            )
            if x.split is None or x.split == 0:
                for i in range(self.n_clusters):
                    samplerange = (
                        x.gshape[0] // self.n_clusters * i,
                        x.gshape[0] // self.n_clusters * (i + 1),
                    )
                    sample = ht.random.randint(samplerange[0], samplerange[1]).item()
                    proc = 0
                    for p in range(x.comm.size):
                        if displ[p] > sample:
                            break
                        proc = p
                    xi = ht.zeros(x.shape[1], dtype=x.dtype)
                    if x.comm.rank == proc:
                        idx = sample - displ[proc]
                        xi = ht.array(x.lloc[idx, :], device=x.device, comm=x.comm)
                    xi.comm.Bcast(xi, root=proc)
                    centroids[i, :] = xi

            else:
                raise NotImplementedError("Not implemented for other splitting-axes")

            self._cluster_centers = centroids

        # directly passed centroids
        elif isinstance(self.init, DNDarray):
            if len(self.init.shape) != 2:
                raise ValueError(
                    "passed centroids need to be two-dimensional, but are {}".format(len(self.init))
                )
            if self.init.shape[0] != self.n_clusters or self.init.shape[1] != x.shape[1]:
                raise ValueError("passed centroids do not match cluster count or data shape")
            self._cluster_centers = self.init.resplit(None)

        # Smart centroid guessing, random sampling with probability weight proportional to distance to existing centroids
        elif self.init == "probability_based":
            if x.split is None or x.split == 0:
                centroids = ht.zeros(
                    (self.n_clusters, x.shape[1]), split=None, device=x.device, comm=x.comm
                )
                sample = ht.random.randint(0, x.shape[0] - 1).item()
                _, displ, _ = x.comm.counts_displs_shape(shape=x.shape, axis=0)
                proc = 0
                for p in range(x.comm.size):
                    if displ[p] > sample:
                        break
                    proc = p
                x0 = ht.zeros(x.shape[1], dtype=x.dtype, device=x.device, comm=x.comm)
                if x.comm.rank == proc:
                    idx = sample - displ[proc]
                    x0 = ht.array(x.lloc[idx, :], device=x.device, comm=x.comm)
                x0.comm.Bcast(x0, root=proc)
                centroids[0, :] = x0
                for i in range(1, self.n_clusters):
                    distances = ht.spatial.distance.cdist(x, centroids, quadratic_expansion=True)
                    D2 = distances.min(axis=1)
                    D2.resplit_(axis=None)
                    prob = D2 / D2.sum()
                    random_position = ht.random.rand().item()
                    sample = 0
                    sum = 0
                    for j in range(len(prob)):
                        if sum > random_position:
                            break
                        sum += prob[j].item()
                        sample = j
                    proc = 0
                    for p in range(x.comm.size):
                        if displ[p] > sample:
                            break
                        proc = p
                    xi = ht.zeros(x.shape[1], dtype=x.dtype)
                    if x.comm.rank == proc:
                        idx = sample - displ[proc]
                        xi = ht.array(x.lloc[idx, :], device=x.device, comm=x.comm)
                    xi.comm.Bcast(xi, root=proc)
                    centroids[i, :] = xi

            else:
                raise NotImplementedError("Not implemented for other splitting-axes")
            self._cluster_centers = centroids

        else:
            raise ValueError(
                'init needs to be one of "random", ht.DNDarray or "kmeans++", but was {}'.format(
                    self.init
                )
            )

    def _assign_to_cluster(self, x: DNDarray):
        """
        Assigns the passed data points to the centroids based on the respective metric

        Parameters
        ----------
        x : DNDarray
            Data points, Shape = (n_samples, n_features)
        """
        # calculate the distance matrix and determine the closest centroid
        distances = self._metric(x, self._cluster_centers)
        matching_centroids = distances.argmin(axis=1, keepdims=True)

        return matching_centroids

    def _update_centroids(self, x: DNDarray, matching_centroids: DNDarray):
        """
        The Update strategy is algorithm specific (e.g. calculate mean of assigned points for kmeans, median for kmedians, etc.)

        Parameters
        ----------
        x : DNDarray
            Input Data
        matching_centroids: DNDarray
            Index array of assigned centroids

        """
        raise NotImplementedError()

    def fit(self, x: DNDarray):
        """
        Computes the centroid of the clustering algorithm to fit the data ``x``. The full pipeline is algorithm specific.

        Parameters
        ----------
        x : DNDarray
            Training instances to cluster. Shape = (n_samples, n_features)

        """
        raise NotImplementedError()

    def predict(self, x: DNDarray):
        """
        Predict the closest cluster each sample in ``x`` belongs to.

        In the vector quantization literature, :func:`cluster_centers_` is called the code book and each value returned by
        predict is the index of the closest code in the code book.

        Parameters
        ----------
        x : DNDarray
            New data to predict. Shape = (n_samples, n_features)
        """
        # input sanitation
        if not isinstance(x, DNDarray):
            raise ValueError(f"input needs to be a ht.DNDarray, but was {type(x)}")

        # determine the centroids
        return self._assign_to_cluster(x)
